# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors

import copy
from pathlib import Path
from typing import Any
from unittest.mock import MagicMock

import pytest

from geneva import DEFAULT_UPLOAD_DIR, connect
from geneva.cluster.builder import default_image
from geneva.manifest import GenevaManifest
from geneva.manifest.builder import (
    CondaManifestBuilder,
    PipManifestBuilder,
    SiteManifestBuilder,
)


@pytest.mark.slow
def test_manifest_upload_location_namespace(tmp_path: Path) -> None:
    """Test manifest zips uploaded to correct location for namespace connections."""
    import pyarrow.fs as pafs
    from lance_namespace import DescribeTableRequest

    # Create a directory namespace connection with a system namespace
    system_ns = "test_system"
    db_root = tmp_path / "db"
    db_root.mkdir()

    # Create zip output directory
    zip_dir = tmp_path / "zips"
    zip_dir.mkdir()

    geneva = connect(
        namespace_impl="dir",
        namespace_properties={"root": str(db_root)},
        system_namespace=[system_ns],
    )

    manifest_def = GenevaManifest(
        name="test-manifest",
        local_zip_output_dir=str(zip_dir),
        skip_site_packages=True,  # Skip to make test faster
        delete_local_zips=False,
        pip=["numpy"],
    )

    # Define the manifest (this should upload zips)
    geneva.define_manifest("test-manifest", manifest_def)

    # Verify the manifest was created and has zips
    manifests = geneva.list_manifests()
    assert len(manifests) == 1
    manifest = manifests[0]
    assert len(manifest.zips) > 0, "Expected manifest to have uploaded zips"

    # Get the manifest table location from namespace
    assert geneva.namespace_client is not None
    from geneva.manifest.mgr import MANIFEST_TABLE_NAME

    table_id = [system_ns, MANIFEST_TABLE_NAME]
    response = geneva.namespace_client.describe_table(DescribeTableRequest(id=table_id))
    manifest_table_location = response.location
    assert manifest_table_location is not None

    # Expected upload directory should be {manifest_table_location}/_geneva_uploads
    expected_upload_dir = f"{manifest_table_location.rstrip('/')}/{DEFAULT_UPLOAD_DIR}"

    # Verify that all uploaded zips are in the expected location
    for zip_path_list in manifest.zips:
        for zip_path in zip_path_list:
            # The zip path should start with the expected upload directory
            assert zip_path.startswith(expected_upload_dir), (
                f"Expected zip to be uploaded to {expected_upload_dir}, "
                f"but got {zip_path}"
            )

            # Verify the file actually exists at that location
            filesystem, path = pafs.FileSystem.from_uri(zip_path)
            file_info = filesystem.get_file_info(path)
            assert file_info.type == pafs.FileType.File, (
                f"Expected file to exist at {zip_path}"
            )


@pytest.mark.slow
def test_manifest_upload_location_local(tmp_path: Path) -> None:
    """Test manifest zips uploaded to correct location for local connections."""
    import pyarrow.fs as pafs

    # Create a local connection
    db_path = tmp_path / "db"

    # Create zip output directory
    zip_dir = tmp_path / "zips"
    zip_dir.mkdir()

    geneva = connect(db_path)

    manifest_def = GenevaManifest(
        name="test-manifest",
        local_zip_output_dir=str(tmp_path / "zips"),
        skip_site_packages=True,  # Skip to make test faster
        delete_local_zips=False,
        pip=["numpy"],
    )

    # Define the manifest (this should upload zips)
    geneva.define_manifest("test-manifest", manifest_def)

    # Verify the manifest was created and has zips
    manifests = geneva.list_manifests()
    assert len(manifests) == 1
    manifest = manifests[0]
    assert len(manifest.zips) > 0, "Expected manifest to have uploaded zips"

    # Verify upload directory is within the manifest table
    from geneva.manifest.mgr import MANIFEST_TABLE_NAME

    expected_upload_dir = (
        f"{str(db_path)}/__system/{MANIFEST_TABLE_NAME}.lance/{DEFAULT_UPLOAD_DIR}"
    )

    # Verify that all uploaded zips are in the expected location
    for zip_path_list in manifest.zips:
        for zip_path in zip_path_list:
            # The zip path should start with the expected upload directory
            assert zip_path.startswith(expected_upload_dir), (
                f"Expected zip to be uploaded to {expected_upload_dir}, "
                f"but got {zip_path}"
            )

            # Verify the file actually exists at that location
            filesystem, path = pafs.FileSystem.from_uri(zip_path)
            file_info = filesystem.get_file_info(path)
            assert file_info.type == pafs.FileType.File, (
                f"Expected file to exist at {zip_path}"
            )


@pytest.mark.slow
def test_manifest_crud(tmp_path: Path) -> None:
    mock_uploader = MagicMock()
    mock_uploader.upload_dir = "/mock/upload/dir"
    mock_uploader._file_exists.return_value = False
    mock_uploader.upload.side_effect = lambda path: f"mock://{path.name}"

    geneva = connect(tmp_path)

    manifest_def = GenevaManifest(
        name="test-manifest-1",
        local_zip_output_dir=str(tmp_path),
        skip_site_packages=False,
        delete_local_zips=False,
        pip=["numpy", "pandas"],
        py_modules=["pyarrow"],
    )

    # upload and create
    geneva.define_manifest("test-manifest-1", manifest_def, uploader=mock_uploader)
    m = geneva.list_manifests()[0]
    _assert_manifest_eq(m.as_dict(), manifest_def.as_dict())

    upload_count = mock_uploader.upload.call_count
    assert upload_count >= 1, "files were not uploaded"

    # update - should update metadata and upload new artifacts
    manifest_def.skip_site_packages = True
    geneva.define_manifest("test-manifest-1", manifest_def, uploader=mock_uploader)
    manifests = geneva.list_manifests()
    assert len(manifests) == 1, "expected single manifest"
    m1 = manifests[0].as_dict()
    m2 = manifest_def.as_dict()
    assert m1["checksum"] != m2["checksum"], "checksum should change"
    _assert_manifest_eq(m1, m2)
    assert mock_uploader.upload.call_count >= upload_count, "files were not uploaded"

    # delete
    geneva.delete_manifest("test-manifest-1")
    assert geneva.list_manifests() == []


def _assert_manifest_eq(m1: dict, m2: dict) -> bool:
    m1 = copy.deepcopy(m1)
    m2 = copy.deepcopy(m2)
    # exclude transient fields from comparison
    for f in {"checksum", "zips"}:
        if f in m1:
            del m1[f]
        if f in m2:
            del m2[f]
    assert m1 == m2, "manifests should match"


def test_manifest_conda_passed_to_ray(monkeypatch: Any) -> None:
    """Test that conda dependencies from manifest are passed to Ray runtime_env."""
    import geneva.runners.ray._mgr as ray_mgr_mod

    ray_init_called: dict[str, Any] = {}

    def mock_ray_init(*args: Any, **kwargs: Any) -> None:
        ray_init_called["args"] = args
        ray_init_called["kwargs"] = kwargs

    monkeypatch.setattr("ray.init", mock_ray_init)
    monkeypatch.setattr("ray.shutdown", lambda: None)
    monkeypatch.setattr("ray.is_initialized", lambda: False)

    # Use a realistic conda structure with nested dependencies
    conda_deps = {
        "dependencies": [
            "python=3.8",
            "pip",
            {"pip": ["requests", "chess"]},
        ],
        "channels": ["conda-forge"],
    }

    manifest = CondaManifestBuilder.create("conda-manifest").conda(conda_deps).build()

    with ray_mgr_mod.ray_cluster(local=True, manifest=manifest):
        pass

    assert "kwargs" in ray_init_called
    runtime_env = ray_init_called["kwargs"]["runtime_env"]
    assert "conda" in runtime_env
    assert runtime_env["conda"] == conda_deps
    assert "pip" not in runtime_env


def test_manifest_requirements_path_passed_to_ray(
    monkeypatch: Any, tmp_path: Path
) -> None:
    """Test that requirements_path from manifest is passed to Ray runtime_env."""
    import geneva.runners.ray._mgr as ray_mgr_mod

    ray_init_called: dict[str, Any] = {}

    def mock_ray_init(*args: Any, **kwargs: Any) -> None:
        ray_init_called["args"] = args
        ray_init_called["kwargs"] = kwargs

    monkeypatch.setattr("ray.init", mock_ray_init)
    monkeypatch.setattr("ray.shutdown", lambda: None)
    monkeypatch.setattr("ray.is_initialized", lambda: False)

    req_file = tmp_path / "requirements.txt"
    req_file.write_text("numpy>=1.0\npandas\n")
    req_path = str(req_file)

    manifest = (
        PipManifestBuilder.create("req-manifest").requirements_path(req_path).build()
    )

    with ray_mgr_mod.ray_cluster(local=True, manifest=manifest):
        pass

    assert "kwargs" in ray_init_called
    runtime_env = ray_init_called["kwargs"]["runtime_env"]
    assert "pip" in runtime_env
    assert runtime_env["pip"] == req_path
    assert "conda" not in runtime_env


def test_manifest_conda_environment_path_passed_to_ray(
    monkeypatch: Any, tmp_path: Path
) -> None:
    """Test that conda_environment_path from manifest is passed to Ray runtime_env."""
    import geneva.runners.ray._mgr as ray_mgr_mod

    ray_init_called: dict[str, Any] = {}

    def mock_ray_init(*args: Any, **kwargs: Any) -> None:
        ray_init_called["args"] = args
        ray_init_called["kwargs"] = kwargs

    monkeypatch.setattr("ray.init", mock_ray_init)
    monkeypatch.setattr("ray.shutdown", lambda: None)
    monkeypatch.setattr("ray.is_initialized", lambda: False)

    env_file = tmp_path / "environment.yml"
    env_file.write_text(
        "name: test-env\n"
        "channels:\n"
        "  - conda-forge\n"
        "dependencies:\n"
        "  - python=3.10\n"
        "  - numpy\n"
    )
    env_path = str(env_file)

    manifest = (
        CondaManifestBuilder.create("conda-env-manifest")
        .conda_environment_path(env_path)
        .build()
    )

    with ray_mgr_mod.ray_cluster(local=True, manifest=manifest):
        pass

    assert "kwargs" in ray_init_called
    runtime_env = ray_init_called["kwargs"]["runtime_env"]
    assert "conda" in runtime_env
    assert runtime_env["conda"] == env_path
    assert "pip" not in runtime_env


def test_make_manifest_conda_none() -> None:
    """Test that _make_manifest handles conda=None from stored data."""
    from datetime import datetime, timezone

    from geneva.manifest.mgr import _make_manifest

    args = {
        "name": "test",
        "version": None,
        "pip": [],
        "py_modules": [],
        "head_image": None,
        "worker_image": None,
        "skip_site_packages": True,
        "delete_local_zips": False,
        "local_zip_output_dir": None,
        "zips": [[]],
        "checksum": None,
        "created_at": datetime.now(timezone.utc),
        "created_by": "test",
        "requirements_path": None,
        "conda": None,
        "conda_environment_path": None,
    }
    manifest = _make_manifest(args)
    assert manifest.conda == {}


# =============================================================================
# Type-Safe Builder Tests
# =============================================================================


class TestPipManifestBuilder:
    """Tests for PipManifestBuilder."""

    def test_no_conda_methods(self) -> None:
        """PipManifestBuilder should not have conda methods."""
        builder = PipManifestBuilder()
        assert not hasattr(builder, "conda")
        assert not hasattr(builder, "conda_environment_path")

    def test_has_pip_methods(self) -> None:
        """PipManifestBuilder should have pip methods."""
        builder = PipManifestBuilder()
        assert hasattr(builder, "pip")
        assert hasattr(builder, "add_pip")
        assert hasattr(builder, "requirements_path")

    def test_pip_packages(self) -> None:
        """Test setting pip packages."""
        manifest = PipManifestBuilder.create("test").pip(["numpy", "pandas"]).build()
        assert manifest.pip == ["numpy", "pandas"]
        assert manifest.conda == {}

    def test_add_pip(self) -> None:
        """Test adding pip packages incrementally."""
        manifest = (
            PipManifestBuilder.create("test").add_pip("numpy").add_pip("pandas").build()
        )
        assert manifest.pip == ["numpy", "pandas"]

    def test_requirements_path(self, tmp_path: Path) -> None:
        """Test using requirements.txt path."""
        req_file = tmp_path / "requirements.txt"
        req_file.write_text("numpy\npandas\n")

        manifest = (
            PipManifestBuilder.create("test").requirements_path(str(req_file)).build()
        )
        assert manifest.requirements_path == str(req_file)
        assert manifest.pip == []

    def test_pip_and_requirements_mutually_exclusive(self) -> None:
        """Test that pip and requirements_path cannot both be set."""
        with pytest.raises(ValueError, match="Cannot set both pip and requirements"):
            PipManifestBuilder.create("test").pip(["numpy"]).requirements_path(
                "/path"
            ).build()

    def test_common_methods(self) -> None:
        """Test that common methods work."""
        manifest = (
            PipManifestBuilder.create("test")
            .version("1.0")
            .py_modules(["mymodule"])
            .head_image("custom:latest")
            .worker_image("worker:latest")
            .upload_site_packages(True)
            .pip(["numpy"])
            .build()
        )
        assert manifest.version == "1.0"
        assert manifest.py_modules == ["mymodule"]
        assert manifest.head_image == "custom:latest"
        assert manifest.worker_image == "worker:latest"
        assert manifest.skip_site_packages is False

    def test_requires_name(self) -> None:
        """Test that builder requires a name."""
        with pytest.raises(ValueError, match="Manifest name is required"):
            PipManifestBuilder().build()


class TestCondaManifestBuilder:
    """Tests for CondaManifestBuilder."""

    def test_no_pip_methods(self) -> None:
        """CondaManifestBuilder should not have pip methods."""
        builder = CondaManifestBuilder()
        assert not hasattr(builder, "pip")
        assert not hasattr(builder, "add_pip")
        assert not hasattr(builder, "requirements_path")

    def test_has_conda_methods(self) -> None:
        """CondaManifestBuilder should have conda methods."""
        builder = CondaManifestBuilder()
        assert hasattr(builder, "conda")
        assert hasattr(builder, "conda_environment_path")

    def test_conda_dependencies(self) -> None:
        """Test setting conda dependencies."""
        conda_deps = {"dependencies": ["python=3.10", "numpy"]}
        manifest = CondaManifestBuilder.create("test").conda(conda_deps).build()
        assert manifest.conda == conda_deps
        assert manifest.pip == []

    def test_conda_environment_path(self, tmp_path: Path) -> None:
        """Test using environment.yml path."""
        env_file = tmp_path / "environment.yml"
        env_file.write_text("name: test\ndependencies:\n  - numpy\n")

        manifest = (
            CondaManifestBuilder.create("test")
            .conda_environment_path(str(env_file))
            .build()
        )
        assert manifest.conda_environment_path == str(env_file)
        assert manifest.conda == {}

    def test_conda_and_path_mutually_exclusive(self) -> None:
        """Test that conda and conda_environment_path cannot both be set."""
        with pytest.raises(ValueError, match="Cannot set both conda and conda_env"):
            CondaManifestBuilder.create("test").conda(
                {"dependencies": ["numpy"]}
            ).conda_environment_path("/path").build()

    def test_common_methods(self) -> None:
        """Test that common methods work."""
        manifest = (
            CondaManifestBuilder.create("test")
            .version("2.0")
            .add_py_module("mymodule")
            .default_head_image()
            .upload_site_packages()
            .conda({"dependencies": ["numpy"]})
            .build()
        )
        assert manifest.version == "2.0"
        assert manifest.py_modules == ["mymodule"]
        assert manifest.head_image == default_image()
        assert manifest.skip_site_packages is False

    def test_requires_name(self) -> None:
        """Test that builder requires a name."""
        with pytest.raises(ValueError, match="Manifest name is required"):
            CondaManifestBuilder().build()


class TestSiteManifestBuilder:
    """Tests for SiteManifestBuilder."""

    def test_no_pip_methods(self) -> None:
        """SiteManifestBuilder should not have pip methods."""
        builder = SiteManifestBuilder()
        assert not hasattr(builder, "pip")
        assert not hasattr(builder, "add_pip")
        assert not hasattr(builder, "requirements_path")

    def test_no_conda_methods(self) -> None:
        """SiteManifestBuilder should not have conda methods."""
        builder = SiteManifestBuilder()
        assert not hasattr(builder, "conda")
        assert not hasattr(builder, "conda_environment_path")

    def test_upload_site_packages_default_true(self) -> None:
        """SiteManifestBuilder should default to uploading site packages."""
        manifest = SiteManifestBuilder.create("test").build()
        assert manifest.skip_site_packages is False  # upload=True means skip=False

    def test_can_disable_site_packages(self) -> None:
        """Test that site packages can be disabled."""
        manifest = (
            SiteManifestBuilder.create("test").upload_site_packages(False).build()
        )
        assert manifest.skip_site_packages is True

    def test_no_external_deps(self) -> None:
        """Test that no external dependencies are set."""
        manifest = SiteManifestBuilder.create("test").build()
        assert manifest.pip == []
        assert manifest.conda == {}
        assert manifest.requirements_path is None
        assert manifest.conda_environment_path is None

    def test_common_methods(self) -> None:
        """Test that common methods work."""
        manifest = (
            SiteManifestBuilder.create("test")
            .version("3.0")
            .py_modules(["local_module"])
            .head_image("site:latest")
            .delete_local_zips(True)
            .build()
        )
        assert manifest.version == "3.0"
        assert manifest.py_modules == ["local_module"]
        assert manifest.head_image == "site:latest"
        assert manifest.delete_local_zips is True

    def test_requires_name(self) -> None:
        """Test that builder requires a name."""
        with pytest.raises(ValueError, match="Manifest name is required"):
            SiteManifestBuilder().build()


class TestGenevaManifestFactories:
    """Tests for GenevaManifest static factory methods."""

    def test_create_pip_factory(self) -> None:
        """GenevaManifest.create_pip() should return PipManifestBuilder."""
        builder = GenevaManifest.create_pip("test")
        assert isinstance(builder, PipManifestBuilder)
        manifest = builder.pip(["numpy"]).build()
        assert manifest.name == "test"
        assert manifest.pip == ["numpy"]

    def test_create_conda_factory(self) -> None:
        """GenevaManifest.create_conda() should return CondaManifestBuilder."""
        builder = GenevaManifest.create_conda("test")
        assert isinstance(builder, CondaManifestBuilder)
        manifest = builder.conda({"dependencies": ["numpy"]}).build()
        assert manifest.name == "test"
        assert manifest.conda == {"dependencies": ["numpy"]}

    def test_create_site_factory(self) -> None:
        """GenevaManifest.create_site() should return SiteManifestBuilder."""
        builder = GenevaManifest.create_site("test")
        assert isinstance(builder, SiteManifestBuilder)
        manifest = builder.build()
        assert manifest.name == "test"
        assert manifest.skip_site_packages is False  # upload enabled by default

    def test_full_workflow_pip(self) -> None:
        """Test complete workflow using pip factory."""
        manifest = (
            GenevaManifest.create_pip("my-pip-manifest")
            .version("1.0.0")
            .pip(["numpy", "pandas"])
            .add_py_module("mymodule")
            .head_image("custom:latest")
            .build()
        )

        assert manifest.name == "my-pip-manifest"
        assert manifest.version == "1.0.0"
        assert manifest.pip == ["numpy", "pandas"]
        assert manifest.py_modules == ["mymodule"]
        assert manifest.head_image == "custom:latest"
        assert manifest.conda == {}

    def test_full_workflow_conda(self) -> None:
        """Test complete workflow using conda factory."""
        conda_deps = {
            "channels": ["conda-forge"],
            "dependencies": ["python=3.10", "numpy", {"pip": ["requests"]}],
        }
        manifest = (
            GenevaManifest.create_conda("my-conda-manifest")
            .conda(conda_deps)
            .upload_site_packages(True)
            .build()
        )

        assert manifest.name == "my-conda-manifest"
        assert manifest.conda == conda_deps
        assert manifest.pip == []
        assert manifest.skip_site_packages is False



# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors

"""Tests for UDTF refresh via ActorPool / JobTracker integration."""

from collections.abc import Iterator

import pyarrow as pa
import pytest

import geneva
from geneva import connect
from geneva.db import Connection
from geneva.table import Table

pytestmark = pytest.mark.ray

OUTPUT_SCHEMA = pa.schema(
    [
        pa.field("id", pa.int64()),
        pa.field("value", pa.float64()),
    ]
)

GROUP_OUTPUT_SCHEMA = pa.schema(
    [
        pa.field("group", pa.string()),
        pa.field("total", pa.float64()),
    ]
)


def _make_source_table(tmp_path, name: str = "source") -> tuple[Connection, Table]:
    db = connect(tmp_path)
    data = pa.table(
        {
            "id": pa.array([1, 2, 3, 4]),
            "value": pa.array([10.0, 20.0, 30.0, 40.0]),
            "group": pa.array(["a", "a", "b", "b"]),
        }
    )
    tbl = db.create_table(name, data)
    return db, tbl


# ---------------------------------------------------------------------------
# ActorPool integration
# ---------------------------------------------------------------------------


class TestUDTFWithActorPool:
    def test_refresh_with_concurrency_param(self, tmp_path, local_ray_context) -> None:
        """Refresh with explicit concurrency=2 produces correct results."""
        db, source_table = _make_source_table(tmp_path)

        @geneva.udtf(output_schema=OUTPUT_SCHEMA, input_columns=["id", "value"])
        def double_values(source) -> Iterator[pa.RecordBatch]:
            tbl = source.to_arrow()
            yield pa.RecordBatch.from_pydict(
                {
                    "id": tbl.column("id").to_pylist(),
                    "value": [v * 2 for v in tbl.column("value").to_pylist()],
                }
            )

        query = source_table.search(None).select(["id", "value"])
        view = db.create_udtf_view("double_view", query, double_values)

        view.refresh(concurrency=2, _admission_check=False)

        result = view.to_arrow()
        assert result.num_rows == 4
        assert sorted(result.column("value").to_pylist()) == [
            20.0,
            40.0,
            60.0,
            80.0,
        ]

    def test_partitioned_refresh_with_concurrency(
        self, tmp_path, local_ray_context
    ) -> None:
        """Partitioned UDTF + ActorPool produces correct aggregation."""
        db, source_table = _make_source_table(tmp_path)

        @geneva.udtf(
            output_schema=GROUP_OUTPUT_SCHEMA,
            input_columns=["group", "value"],
            partition_by="group",
        )
        def sum_by_group(source) -> Iterator[pa.RecordBatch]:
            tbl = source.to_arrow()
            total = sum(tbl.column("value").to_pylist())
            group_val = tbl.column("group")[0].as_py()
            yield pa.RecordBatch.from_pydict(
                {
                    "group": [group_val],
                    "total": [total],
                }
            )

        query = source_table.search(None).select(["group", "value"])
        view = db.create_udtf_view("grouped_view", query, sum_by_group)

        view.refresh(concurrency=2, _admission_check=False)

        result = view.to_arrow()
        assert result.num_rows == 2
        rows = result.to_pydict()
        by_group = dict(zip(rows["group"], rows["total"], strict=True))
        assert by_group["a"] == 30.0  # 10 + 20
        assert by_group["b"] == 70.0  # 30 + 40

    def test_concurrency_capped_to_num_partitions(
        self, tmp_path, local_ray_context
    ) -> None:
        """concurrency=100 with only 2 partitions still works correctly."""
        db, source_table = _make_source_table(tmp_path)

        @geneva.udtf(
            output_schema=GROUP_OUTPUT_SCHEMA,
            input_columns=["group", "value"],
            partition_by="group",
        )
        def sum_by_group(source) -> Iterator[pa.RecordBatch]:
            tbl = source.to_arrow()
            total = sum(tbl.column("value").to_pylist())
            group_val = tbl.column("group")[0].as_py()
            yield pa.RecordBatch.from_pydict(
                {
                    "group": [group_val],
                    "total": [total],
                }
            )

        query = source_table.search(None).select(["group", "value"])
        view = db.create_udtf_view("capped_view", query, sum_by_group)

        # concurrency >> partitions — should cap to 2 actors
        view.refresh(concurrency=100, _admission_check=False)

        result = view.to_arrow()
        assert result.num_rows == 2


# ---------------------------------------------------------------------------
# Admission control
# ---------------------------------------------------------------------------


class TestUDTFAdmissionControl:
    def test_admission_check_false_skips_check(
        self, tmp_path, local_ray_context
    ) -> None:
        """_admission_check=False allows refresh without admission validation."""
        db, source_table = _make_source_table(tmp_path)

        @geneva.udtf(output_schema=OUTPUT_SCHEMA, input_columns=["id", "value"])
        def passthrough(source) -> Iterator[pa.RecordBatch]:
            yield from source.to_batches()

        query = source_table.search(None).select(["id", "value"])
        view = db.create_udtf_view("admission_view", query, passthrough)

        # Should not raise even though we're on a small local cluster
        view.refresh(_admission_check=False)

        assert view.count_rows() == 4


# ---------------------------------------------------------------------------
# on_error handling
# ---------------------------------------------------------------------------

# Module-level counter for retry tests (must be serialisable by cloudpickle)
_retry_call_count: int = 0


class TestUDTFOnError:
    def test_default_fail_raises_on_udtf_exception(
        self, tmp_path, local_ray_context
    ) -> None:
        """No on_error configured → exception propagates and refresh fails."""
        db, source_table = _make_source_table(tmp_path)

        @geneva.udtf(output_schema=OUTPUT_SCHEMA, input_columns=["id", "value"])
        def failing_udtf(source) -> Iterator[pa.RecordBatch]:
            raise ValueError("boom")

        query = source_table.search(None).select(["id", "value"])
        view = db.create_udtf_view("fail_view", query, failing_udtf)

        with pytest.raises(ValueError, match="boom"):
            view.refresh(concurrency=1, _admission_check=False)

    def test_skip_continues_on_partition_failure(
        self, tmp_path, local_ray_context
    ) -> None:
        """on_error=[Skip(ValueError)] skips the failing partition, keeps others."""
        db, source_table = _make_source_table(tmp_path)

        @geneva.udtf(
            output_schema=GROUP_OUTPUT_SCHEMA,
            input_columns=["group", "value"],
            partition_by="group",
            on_error=[geneva.Skip(ValueError)],
        )
        def skip_udtf(source) -> Iterator[pa.RecordBatch]:
            tbl = source.to_arrow()
            group_val = tbl.column("group")[0].as_py()
            if group_val == "a":
                raise ValueError("partition a fails")
            total = sum(tbl.column("value").to_pylist())
            yield pa.RecordBatch.from_pydict({"group": [group_val], "total": [total]})

        query = source_table.search(None).select(["group", "value"])
        view = db.create_udtf_view("skip_view", query, skip_udtf)
        view.refresh(concurrency=2, _admission_check=False)

        result = view.to_arrow()
        # Only partition "b" succeeded
        assert result.num_rows == 1
        rows = result.to_pydict()
        assert rows["group"] == ["b"]
        assert rows["total"] == [70.0]

    def test_retry_succeeds_after_transient_failure(
        self, tmp_path, local_ray_context
    ) -> None:
        """on_error=[Retry(ValueError, max_attempts=3)] retries and succeeds."""
        db, source_table = _make_source_table(tmp_path)

        # Use a mutable container so the closure inside the UDTF (which runs
        # in a Ray actor) can track how many times it has been called.
        call_counts: dict[str, int] = {}

        @geneva.udtf(
            output_schema=OUTPUT_SCHEMA,
            input_columns=["id", "value"],
            on_error=[geneva.Retry(ValueError, max_attempts=3)],
        )
        def retry_udtf(source) -> Iterator[pa.RecordBatch]:
            # Fail on the first call, succeed on the second
            key = "calls"
            call_counts[key] = call_counts.get(key, 0) + 1
            if call_counts[key] < 2:
                raise ValueError("transient failure")
            tbl = source.to_arrow()
            yield pa.RecordBatch.from_pydict(
                {
                    "id": tbl.column("id").to_pylist(),
                    "value": tbl.column("value").to_pylist(),
                }
            )

        query = source_table.search(None).select(["id", "value"])
        view = db.create_udtf_view("retry_view", query, retry_udtf)
        view.refresh(concurrency=1, _admission_check=False)

        result = view.to_arrow()
        assert result.num_rows == 4

    def test_retry_exhausted_raises(self, tmp_path, local_ray_context) -> None:
        """Retries exhausted → exception propagates."""
        db, source_table = _make_source_table(tmp_path)

        @geneva.udtf(
            output_schema=OUTPUT_SCHEMA,
            input_columns=["id", "value"],
            on_error=[geneva.Retry(ValueError, max_attempts=2)],
        )
        def always_fails(source) -> Iterator[pa.RecordBatch]:
            raise ValueError("permanent failure")

        query = source_table.search(None).select(["id", "value"])
        view = db.create_udtf_view("exhaust_view", query, always_fails)

        with pytest.raises(ValueError, match="permanent failure"):
            view.refresh(concurrency=1, _admission_check=False)


# ---------------------------------------------------------------------------
# ErrorStore integration
# ---------------------------------------------------------------------------


class TestUDTFErrorStoreIntegration:
    def test_skip_logs_error_to_error_store(self, tmp_path, local_ray_context) -> None:
        """Skip(ValueError) persists an ErrorRecord to ErrorStore."""
        db, source_table = _make_source_table(tmp_path)

        @geneva.udtf(
            output_schema=GROUP_OUTPUT_SCHEMA,
            input_columns=["group", "value"],
            partition_by="group",
            on_error=[geneva.Skip(ValueError)],
        )
        def skip_udtf(source) -> Iterator[pa.RecordBatch]:
            tbl = source.to_arrow()
            group_val = tbl.column("group")[0].as_py()
            if group_val == "a":
                raise ValueError("partition a fails")
            total = sum(tbl.column("value").to_pylist())
            yield pa.RecordBatch.from_pydict({"group": [group_val], "total": [total]})

        query = source_table.search(None).select(["group", "value"])
        view = db.create_udtf_view("skip_err_view", query, skip_udtf)
        view.refresh(concurrency=2, _admission_check=False)

        # Verify ErrorStore contains the error
        from geneva.debug.error_store import ErrorStore

        error_store = ErrorStore(db, namespace=db.system_namespace)
        errors = error_store.get_errors(table_name=view.name)
        assert len(errors) == 1
        err = errors[0]
        assert err.error_type == "ValueError"
        assert "partition a fails" in err.error_message
        assert err.table_name == view.name
        assert err.udf_name == "skip_udtf"


# ---------------------------------------------------------------------------
# Checkpoint-based assembly
# ---------------------------------------------------------------------------


class TestUDTFCheckpointAssembly:
    def test_checkpoint_assembly_does_not_duplicate_data(
        self, tmp_path, local_ray_context
    ) -> None:
        """Multi-partition UDTF: row count matches expected (no duplication)."""
        db, source_table = _make_source_table(tmp_path)

        @geneva.udtf(
            output_schema=GROUP_OUTPUT_SCHEMA,
            input_columns=["group", "value"],
            partition_by="group",
        )
        def sum_by_group(source) -> Iterator[pa.RecordBatch]:
            tbl = source.to_arrow()
            total = sum(tbl.column("value").to_pylist())
            group_val = tbl.column("group")[0].as_py()
            yield pa.RecordBatch.from_pydict({"group": [group_val], "total": [total]})

        query = source_table.search(None).select(["group", "value"])
        view = db.create_udtf_view("ckp_view", query, sum_by_group)

        view.refresh(concurrency=2, _admission_check=False)

        result = view.to_arrow()
        # 2 partitions → 2 rows, no duplication from checkpoint reads
        assert result.num_rows == 2
        rows = result.to_pydict()
        by_group = dict(zip(rows["group"], rows["total"], strict=True))
        assert by_group["a"] == 30.0
        assert by_group["b"] == 70.0

    def test_partitioned_refresh_creates_multiple_fragments(
        self, tmp_path, local_ray_context
    ) -> None:
        """Each partition should be written as its own Lance fragment."""
        import lance

        db, source_table = _make_source_table(tmp_path)

        @geneva.udtf(
            output_schema=GROUP_OUTPUT_SCHEMA,
            input_columns=["group", "value"],
            partition_by="group",
        )
        def sum_by_group_frag(source) -> Iterator[pa.RecordBatch]:
            tbl = source.to_arrow()
            total = sum(tbl.column("value").to_pylist())
            group_val = tbl.column("group")[0].as_py()
            yield pa.RecordBatch.from_pydict({"group": [group_val], "total": [total]})

        query = source_table.search(None).select(["group", "value"])
        view = db.create_udtf_view("frag_view", query, sum_by_group_frag)

        view.refresh(concurrency=2, _admission_check=False)

        # Verify correctness
        result = view.to_arrow()
        assert result.num_rows == 2

        # Verify that the dataset has one fragment per partition
        ds = lance.dataset(view.uri)
        fragments = ds.get_fragments()
        assert len(fragments) == 2



# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors

"""Output formatters for CLI responses."""

import json
from datetime import datetime, timezone
from enum import Enum

from .models import Cluster, Job, Manifest, RayCluster
from rich.console import Console
from rich.table import Table
from rich.text import Text


class OutputFormat(str, Enum):
    """Output format enum."""

    TEXT = "text"
    JSON = "json"
    TABLE = "table"


console = Console()


def _format_datetime(dt: datetime | None) -> str:
    """Format datetime for display."""
    if dt is None:
        return "-"

    now = datetime.now(timezone.utc)
    if dt.tzinfo is None:
        dt = dt.replace(tzinfo=timezone.utc)

    delta = now - dt
    seconds = delta.total_seconds()

    if seconds < 60:
        return f"{int(seconds)}s ago"
    elif seconds < 3600:
        return f"{int(seconds / 60)}m ago"
    elif seconds < 86400:
        return f"{int(seconds / 3600)}h ago"
    elif seconds < 604800:
        return f"{int(seconds / 86400)}d ago"
    else:
        return dt.strftime("%Y-%m-%d %H:%M")


def _status_style(status: str) -> str:
    """Get rich style for job status."""
    status_lower = status.lower()
    if status_lower in ("completed", "done", "ready"):
        return "green"
    elif status_lower == "running":
        return "green bold"
    elif status_lower in ("failed", "error"):
        return "red bold"
    elif status_lower in ("pending", "initializing"):
        return "yellow"
    elif status_lower == "cancelled":
        return "dim"
    return "white"


def _format_duration(
    launched_at: datetime | None, completed_at: datetime | None
) -> str:
    """Calculate and format job duration."""
    if not launched_at:
        return "-"

    end_time = completed_at or datetime.now(timezone.utc)
    if launched_at.tzinfo is None:
        launched_at = launched_at.replace(tzinfo=timezone.utc)
    if end_time.tzinfo is None:
        end_time = end_time.replace(tzinfo=timezone.utc)
    delta = end_time - launched_at
    total_seconds = int(delta.total_seconds())

    if total_seconds < 60:
        return f"{total_seconds}s"
    elif total_seconds < 3600:
        return f"{total_seconds // 60}m {total_seconds % 60}s"
    else:
        hours = total_seconds // 3600
        mins = (total_seconds % 3600) // 60
        return f"{hours}h {mins}m"


def format_clusters(clusters: list[Cluster], fmt: OutputFormat) -> None:
    """Format and print cluster list."""
    if fmt == OutputFormat.JSON:
        data = [c.model_dump(by_alias=True) for c in clusters]
        console.print_json(json.dumps(data, default=str))
        return

    if not clusters:
        console.print("[dim]No clusters found[/dim]")
        return

    if fmt == OutputFormat.TABLE:
        table = Table(show_header=True, header_style="bold cyan")
        table.add_column("Name", style="bold")
        table.add_column("Type")
        table.add_column("Created By")
        table.add_column("Created")

        for cluster in clusters:
            table.add_row(
                cluster.name,
                cluster.cluster_type or "-",
                cluster.created_by or "-",
                _format_datetime(cluster.created_at),
            )
        console.print(table)
    else:
        # Text format
        for cluster in clusters:
            console.print(f"[bold]{cluster.name}[/bold]")
            if cluster.cluster_type:
                console.print(f"  [dim]Type:[/dim] {cluster.cluster_type}")
            if cluster.created_by:
                console.print(f"  [dim]Created By:[/dim] {cluster.created_by}")
            if cluster.created_at:
                console.print(
                    f"  [dim]Created:[/dim] {_format_datetime(cluster.created_at)}"
                )


def format_manifests(manifests: list[Manifest], fmt: OutputFormat) -> None:
    """Format and print manifest list."""
    if fmt == OutputFormat.JSON:
        data = [m.model_dump(by_alias=True) for m in manifests]
        console.print_json(json.dumps(data, default=str))
        return

    if not manifests:
        console.print("[dim]No manifests found[/dim]")
        return

    if fmt == OutputFormat.TABLE:
        table = Table(show_header=True, header_style="bold cyan")
        table.add_column("Name", style="bold")
        table.add_column("Version")
        table.add_column("Checksum", max_width=12)
        table.add_column("Created By")
        table.add_column("Created")

        for manifest in manifests:
            checksum = manifest.checksum[:8] + "..." if manifest.checksum else "-"
            table.add_row(
                manifest.name,
                manifest.version or "-",
                checksum,
                manifest.created_by or "-",
                _format_datetime(manifest.created_at),
            )
        console.print(table)
    else:
        # Text format
        for manifest in manifests:
            version = f" v{manifest.version}" if manifest.version else ""
            console.print(f"[bold]{manifest.name}[/bold][dim]{version}[/dim]")
            if manifest.created_by:
                console.print(f"  [dim]Created By:[/dim] {manifest.created_by}")
            if manifest.created_at:
                console.print(
                    f"  [dim]Created:[/dim] {_format_datetime(manifest.created_at)}"
                )


def format_jobs(jobs: list[Job], fmt: OutputFormat) -> None:
    """Format and print job list."""
    if fmt == OutputFormat.JSON:
        data = [j.model_dump(by_alias=True) for j in jobs]
        console.print_json(json.dumps(data, default=str))
        return

    if not jobs:
        console.print("[dim]No jobs found[/dim]")
        return

    if fmt == OutputFormat.TABLE:
        table = Table(show_header=True, header_style="bold cyan")
        table.add_column("Job ID", style="bold", max_width=36)
        table.add_column("Type")
        table.add_column("Table")
        table.add_column("Status")
        table.add_column("Launched By")
        table.add_column("Launched")
        table.add_column("Duration")

        for job in jobs:
            # Calculate duration
            duration = _format_duration(job.launched_at, job.completed_at)

            # Truncate table name for display
            table_name = job.table_name or "-"
            if len(table_name) > 16:
                table_name = table_name[:13] + "..."

            table.add_row(
                job.job_id,
                job.job_type or "-",
                table_name,
                Text(job.status.value, style=_status_style(job.status.value)),
                job.launched_by or "-",
                _format_datetime(job.launched_at),
                duration,
            )
        console.print(table)
    else:
        # Text format
        for job in jobs:
            status_style = _status_style(job.status.value)
            console.print(
                f"[bold]{job.job_id}[/bold] [{status_style}]{job.status.value}[/]"
            )
            if job.job_type:
                console.print(f"  [dim]Type:[/dim] {job.job_type}")
            if job.table_name:
                console.print(f"  [dim]Table:[/dim] {job.table_name}")
            if job.column_name:
                console.print(f"  [dim]Column:[/dim] {job.column_name}")
            if job.launched_by:
                console.print(f"  [dim]Launched By:[/dim] {job.launched_by}")
            if job.launched_at:
                console.print(
                    f"  [dim]Launched:[/dim] {_format_datetime(job.launched_at)}"
                )


def format_ray_clusters(clusters: list[RayCluster], fmt: OutputFormat) -> None:
    """Format and print RayCluster list."""
    if fmt == OutputFormat.JSON:
        data = [c.model_dump() for c in clusters]
        console.print_json(json.dumps(data, default=str))
        return

    if not clusters:
        console.print("[dim]No RayClusters found[/dim]")
        return

    if fmt == OutputFormat.TABLE:
        table = Table(show_header=True, header_style="bold cyan")
        table.add_column("Name", style="bold")
        table.add_column("Namespace")
        table.add_column("Status")
        table.add_column("Head Pod")
        table.add_column("Workers")
        table.add_column("Created")

        for cluster in clusters:
            status = cluster.status or "-"
            table.add_row(
                cluster.name,
                cluster.namespace,
                Text(status, style=_status_style(status)),
                cluster.head_pod or "-",
                str(cluster.worker_replicas)
                if cluster.worker_replicas is not None
                else "-",
                _format_datetime(cluster.created_at),
            )
        console.print(table)
    else:
        # Text format
        for cluster in clusters:
            status = cluster.status or ""
            status_text = f" [{_status_style(status)}]{status}[/]" if status else ""
            console.print(f"[bold]{cluster.name}[/bold]{status_text}")
            console.print(f"  [dim]Namespace:[/dim] {cluster.namespace}")
            if cluster.head_pod:
                console.print(f"  [dim]Head:[/dim] {cluster.head_pod}")
            if cluster.worker_replicas is not None:
                console.print(f"  [dim]Workers:[/dim] {cluster.worker_replicas}")
            if cluster.created_at:
                console.print(
                    f"  [dim]Created:[/dim] {_format_datetime(cluster.created_at)}"
                )


def format_ray_cluster_detail(cluster: RayCluster, fmt: OutputFormat) -> None:
    """Format and print detailed RayCluster info."""
    if fmt == OutputFormat.JSON:
        console.print_json(json.dumps(cluster.model_dump(), default=str))
        return

    status = cluster.status or ""
    status_text = f" [{_status_style(status)}]{status}[/]" if status else ""
    console.print(
        f"[bold cyan]RayCluster:[/bold cyan] [bold]{cluster.name}[/bold]{status_text}"
    )
    console.print(f"  [dim]Namespace:[/dim]   {cluster.namespace}")
    if cluster.head_pod:
        console.print(f"  [dim]Head Pod:[/dim]    {cluster.head_pod}")
    if cluster.worker_replicas is not None:
        console.print(f"  [dim]Workers:[/dim]     {cluster.worker_replicas}")
    console.print(f"  [dim]Created:[/dim]     {_format_datetime(cluster.created_at)}")


# ============================================================================
# Detailed formatters for single resources (get commands)
# ============================================================================


def format_cluster_detail(cluster: Cluster, fmt: OutputFormat) -> None:
    """Format and print detailed cluster info."""
    if fmt == OutputFormat.JSON:
        data = cluster.model_dump(by_alias=True)
        console.print_json(json.dumps(data, default=str))
        return

    console.print(f"[bold cyan]Cluster:[/bold cyan] [bold]{cluster.name}[/bold]")
    if cluster.cluster_type:
        console.print(f"  [dim]Type:[/dim]        {cluster.cluster_type}")
    if cluster.ray_address:
        console.print(f"  [dim]Ray Address:[/dim] {cluster.ray_address}")
    if cluster.created_by:
        console.print(f"  [dim]Created By:[/dim]  {cluster.created_by}")
    console.print(f"  [dim]Created:[/dim]     {_format_datetime(cluster.created_at)}")

    # Show kuberay config if present
    if cluster.kuberay:
        console.print("\n[bold cyan]KubeRay Config:[/bold cyan]")
        kuberay = cluster.kuberay
        if kuberay.get("namespace"):
            console.print(f"  [dim]Namespace:[/dim]      {kuberay['namespace']}")
        if kuberay.get("config_method"):
            console.print(f"  [dim]Config Method:[/dim]  {kuberay['config_method']}")
        if kuberay.get("aws_region"):
            console.print(f"  [dim]AWS Region:[/dim]     {kuberay['aws_region']}")

        # Head group
        if kuberay.get("head_group"):
            hg = kuberay["head_group"]
            console.print("\n  [bold]Head Group:[/bold]")
            if hg.get("image"):
                console.print(f"    [dim]Image:[/dim]    {hg['image']}")
            if hg.get("num_cpus"):
                console.print(f"    [dim]CPUs:[/dim]     {hg['num_cpus']}")
            if hg.get("memory"):
                console.print(f"    [dim]Memory:[/dim]   {hg['memory']}")

        # Worker groups
        if kuberay.get("worker_groups"):
            for i, wg in enumerate(kuberay["worker_groups"]):
                console.print(f"\n  [bold]Worker Group {i + 1}:[/bold]")
                if wg.get("image"):
                    console.print(f"    [dim]Image:[/dim]       {wg['image']}")
                if wg.get("num_cpus"):
                    console.print(f"    [dim]CPUs:[/dim]        {wg['num_cpus']}")
                if wg.get("memory"):
                    console.print(f"    [dim]Memory:[/dim]      {wg['memory']}")
                if wg.get("replicas") is not None:
                    console.print(f"    [dim]Replicas:[/dim]    {wg['replicas']}")
                if wg.get("min_replicas") is not None:
                    console.print(f"    [dim]Min Replicas:[/dim] {wg['min_replicas']}")
                if wg.get("max_replicas") is not None:
                    console.print(f"    [dim]Max Replicas:[/dim] {wg['max_replicas']}")


def format_manifest_detail(manifest: Manifest, fmt: OutputFormat) -> None:
    """Format and print detailed manifest info."""
    if fmt == OutputFormat.JSON:
        data = manifest.model_dump(by_alias=True)
        console.print_json(json.dumps(data, default=str))
        return

    console.print(f"[bold cyan]Manifest:[/bold cyan] [bold]{manifest.name}[/bold]")
    console.print(f"  [dim]Version:[/dim]            {manifest.version or '-'}")
    if manifest.checksum:
        console.print(f"  [dim]Checksum:[/dim]           {manifest.checksum}")
    if manifest.created_by:
        console.print(f"  [dim]Created By:[/dim]         {manifest.created_by}")
    console.print(
        f"  [dim]Created:[/dim]            {_format_datetime(manifest.created_at)}"
    )
    console.print(f"  [dim]Skip Site Packages:[/dim] {manifest.skip_site_packages}")

    # Images
    if manifest.head_image:
        console.print(f"  [dim]Head Image:[/dim]         {manifest.head_image}")
    if manifest.worker_image:
        console.print(f"  [dim]Worker Image:[/dim]       {manifest.worker_image}")

    if manifest.pip:
        console.print("\n[bold cyan]Pip Dependencies:[/bold cyan]")
        for dep in manifest.pip:
            console.print(f"  - {dep}")

    if manifest.py_modules:
        console.print("\n[bold cyan]Python Modules:[/bold cyan]")
        for mod in manifest.py_modules:
            console.print(f"  - {mod}")

    if manifest.zips:
        # Filter out empty zip entries
        non_empty_zips = [z for z in manifest.zips if z]
        if non_empty_zips:
            console.print("\n[bold cyan]Zips:[/bold cyan]")
            for zip_entry in non_empty_zips:
                for z in zip_entry:
                    console.print(f"  - {z}")


def format_job_detail(job: Job, fmt: OutputFormat) -> None:
    """Format and print detailed job info."""
    if fmt == OutputFormat.JSON:
        console.print_json(json.dumps(job.model_dump(by_alias=True), default=str))
        return

    status_style = _status_style(job.status.value)
    console.print(
        f"[bold cyan]Job:[/bold cyan] [bold]{job.job_id}[/bold]"
        f" [{status_style}]{job.status.value}[/]"
    )

    # Basic info
    if job.job_type:
        console.print(f"  [dim]Type:[/dim]             {job.job_type}")
    if job.table_name:
        console.print(f"  [dim]Table:[/dim]            {job.table_name}")
    if job.column_name:
        console.print(f"  [dim]Column:[/dim]           {job.column_name}")

    # Execution info
    if job.manifest_id:
        console.print(f"  [dim]Manifest ID:[/dim]      {job.manifest_id}")
    if job.manifest_checksum:
        console.print(f"  [dim]Manifest Checksum:[/dim] {job.manifest_checksum}")
    if job.launched_by:
        console.print(f"  [dim]Launched By:[/dim]      {job.launched_by}")

    # Timing info
    console.print(f"  [dim]Launched:[/dim]         {_format_datetime(job.launched_at)}")
    if job.completed_at:
        console.print(
            f"  [dim]Completed:[/dim]        {_format_datetime(job.completed_at)}"
        )
    duration = _format_duration(job.launched_at, job.completed_at)
    console.print(f"  [dim]Duration:[/dim]         {duration}")
    if job.updated_at:
        console.print(
            f"  [dim]Updated:[/dim]          {_format_datetime(job.updated_at)}"
        )

    # Config (parse from JSON string if needed)
    config_dict = job.get_config_dict()
    if config_dict:
        console.print("\n[bold cyan]Config:[/bold cyan]")
        for key, value in config_dict.items():
            if value is not None:
                console.print(f"  [dim]{key}:[/dim] {value}")

    # Events (formatted with timeline)
    if job.events:
        console.print("\n[bold cyan]Events:[/bold cyan]")
        for i, event in enumerate(job.events):
            prefix = "[green]✔[/green]" if i == len(job.events) - 1 else "[dim]│[/dim]"
            console.print(f"  {prefix} {event}")

    # Metrics (formatted as table if present)
    parsed_metrics = job.get_metrics_parsed()
    if parsed_metrics:
        console.print("\n[bold cyan]Metrics:[/bold cyan]")
        table = Table(show_header=True, header_style="bold", box=None, padding=(0, 2))
        table.add_column("Name")
        table.add_column("Progress", justify="right")
        table.add_column("Done")
        table.add_column("Description")
        for metric in parsed_metrics:
            progress = (
                f"{metric.n}/{metric.total}" if metric.total > 0 else str(metric.n)
            )
            done_text = Text(
                "✔" if metric.done else "-", style="green" if metric.done else "dim"
            )
            table.add_row(metric.name, progress, done_text, metric.desc)
        console.print(table)
    elif job.metrics:
        # Fallback for unparsed metrics
        console.print("\n[bold cyan]Metrics:[/bold cyan]")
        for metric in job.metrics:
            if metric:
                console.print(f"  - {metric}")



# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors
from pathlib import Path
from unittest.mock import MagicMock, patch

import lancedb
import pyarrow as pa
import pytest

from geneva import connect
from geneva.db import (
    SYSTEM_TABLE_COMPLETED,
    SYSTEM_TABLE_MIGRATING,
    SYSTEM_TABLE_MIGRATIONS_TABLE_NAME,
    Connection,
    _ensure_system_namespace_exists,
)
from geneva.debug.error_store import GENEVA_ERRORS_TABLE_NAME, ErrorStore
from geneva.jobs.jobs import GENEVA_JOBS_TABLE_NAME, JobStateManager
from geneva.manifest.mgr import ManifestConfigManager
from geneva.table import TableReference


def test_connect(tmp_path: Path) -> None:
    db = connect(tmp_path)

    # Use lancedb to verify the results are the same
    ldb = lancedb.connect(tmp_path)

    # Create a Table with integer columns
    tbl = pa.Table.from_pydict({"a": [1, 2, 3], "b": [4, 5, 6]})
    db.create_table("table1", tbl)
    ldb_tbls = db.table_names()
    assert "table1" in ldb_tbls
    db.open_table("table1")

    db_tbls = db.table_names()
    assert db_tbls == ldb_tbls

    # Use lancedb to read the data back
    ldb_tbl = ldb.open_table("table1")
    assert ldb_tbl.to_arrow() == tbl
    db.drop_table("table1")


def test_connect_non_existent(tmp_path: Path) -> None:
    db = connect(tmp_path)
    with pytest.raises(ValueError, match=r".*was not found"):
        db.open_table("non_existent")


def test_error_store_from_system_table_ref_stays_in_single_system_db(
    tmp_path: Path,
) -> None:
    db = connect(tmp_path)
    db.create_table("table1", pa.Table.from_pydict({"a": [1]}))

    system_ref = (
        db.open_table("table1")
        .get_reference()
        .as_system_table(GENEVA_ERRORS_TABLE_NAME)
    )
    reopened = system_ref.open_db()
    store = ErrorStore(reopened)

    assert store.table.uri.endswith("/__system/geneva_errors.lance")
    assert "/__system/__system/" not in store.table.uri


def test_error_store_nests_under_system_db_named_system(tmp_path: Path) -> None:
    db_path = tmp_path / "__system"
    db = connect(db_path)
    db.create_table("table1", pa.Table.from_pydict({"a": [1]}))

    system_ref = (
        db.open_table("table1")
        .get_reference()
        .as_system_table(GENEVA_ERRORS_TABLE_NAME)
    )
    reopened = system_ref.open_db()
    store = ErrorStore(reopened)

    assert store.table.uri.endswith("/__system/__system/geneva_errors.lance")


def test_legacy_system_table_migration_writes_completed_marker(tmp_path: Path) -> None:
    db = connect(tmp_path)
    legacy_jobs = JobStateManager(db, namespace=[])
    legacy_jobs.launch("tbl", "col")

    migrated_jobs = JobStateManager(db)

    assert migrated_jobs.get_table().count_rows() == 1
    assert db.open_table(GENEVA_JOBS_TABLE_NAME).count_rows() == 1

    migrations = db.system_table_connection.open_table(
        SYSTEM_TABLE_MIGRATIONS_TABLE_NAME
    )
    rows = (
        migrations.search()
        .where(f"table_name = '{GENEVA_JOBS_TABLE_NAME}'")
        .limit(1)
        .to_arrow()
        .to_pylist()
    )
    assert rows[0]["status"] == SYSTEM_TABLE_COMPLETED


def test_migrate_legacy_system_table_retries_after_failure(tmp_path: Path) -> None:
    db = connect(tmp_path)
    target_db = db.system_table_connection
    copy_calls = 0
    status_updates: list[str] = []

    def fake_copy(**_: object) -> None:
        nonlocal copy_calls
        copy_calls += 1
        if copy_calls == 1:
            raise RuntimeError("copy failed")

    with (
        patch.object(
            Connection, "_get_system_table_migration_status", return_value=None
        ),
        patch.object(
            Connection,
            "_find_legacy_system_table_source",
            return_value=(db, []),
        ),
        patch.object(
            Connection,
            "_set_system_table_migration_status",
            side_effect=lambda table_name, status: status_updates.append(status),
        ),
        patch.object(Connection, "_replace_table_from_source", side_effect=fake_copy),
    ):
        with pytest.raises(RuntimeError, match="copy failed"):
            db.migrate_legacy_system_table_if_needed(
                table_name=GENEVA_JOBS_TABLE_NAME,
                target_conn=target_db,
                target_namespace=[],
            )

        db.migrate_legacy_system_table_if_needed(
            table_name=GENEVA_JOBS_TABLE_NAME,
            target_conn=target_db,
            target_namespace=[],
        )

    assert copy_calls == 2


def test_migrate_legacy_system_table_clears_stale_marker_without_legacy_source(
    tmp_path: Path,
) -> None:
    db = connect(tmp_path)
    target_db = db.system_table_connection

    with (
        patch.object(
            Connection,
            "_get_system_table_migration_status",
            return_value=SYSTEM_TABLE_MIGRATING,
        ),
        patch.object(
            Connection,
            "_find_legacy_system_table_source",
            return_value=None,
        ),
        patch.object(Connection, "_clear_system_table_migration_status") as mock_clear,
        patch.object(Connection, "_replace_table_from_source") as mock_copy,
    ):
        db.migrate_legacy_system_table_if_needed(
            table_name=GENEVA_JOBS_TABLE_NAME,
            target_conn=target_db,
            target_namespace=[],
        )

    mock_clear.assert_called_once_with(GENEVA_JOBS_TABLE_NAME)
    mock_copy.assert_not_called()


class TestConnectRemote:
    """Tests for connect() with remote db:// URI (Phalanx server)."""

    @pytest.fixture(autouse=True)
    def _mock_set_upload_dir(self) -> None:
        with patch("geneva.db._set_upload_dir"):
            yield

    def test_connect_remote_uri_uses_host_override(self) -> None:
        """Test that db:// URI with host_override triggers remote connection."""
        with patch("geneva.db.lancedb.connect") as mock_lancedb_connect:
            mock_conn = MagicMock()
            mock_conn.table_names.return_value = []
            mock_lancedb_connect.return_value = mock_conn

            db = connect(
                uri="db://my_database",
                api_key="test-api-key",
                host_override="https://phalanx.internal:8080",
            )

            # Trigger the lazy connection
            _ = db._connect

            # Verify lancedb.connect was called with correct parameters
            mock_lancedb_connect.assert_called_once()
            call_kwargs = mock_lancedb_connect.call_args.kwargs
            assert call_kwargs.get("api_key") is not None
            assert call_kwargs.get("host_override") == "https://phalanx.internal:8080"

            # Verify the URI was passed correctly
            call_args = mock_lancedb_connect.call_args.args
            assert call_args[0] == "db://my_database"

    def test_connect_remote_stores_connection_params(self) -> None:
        """Test that remote connection parameters are stored on Connection object."""
        db = connect(
            uri="db://test_db",
            api_key="my-api-key",
            host_override="https://phalanx.example.com",
            region="us-west-2",
        )

        assert db.uri == "db://test_db"
        assert db._host_override == "https://phalanx.example.com"
        assert db._region == "us-west-2"
        # api_key is wrapped in Credential
        assert db._api_key is not None

    def test_connect_remote_system_namespace_preserved(self) -> None:
        """Remote connections should preserve configured system_namespace."""
        db = connect(
            uri="db://test_db",
            api_key="test-key",
            host_override="https://phalanx.example.com",
            system_namespace=["prod", "system"],
        )

        assert db.system_namespace == ["prod", "system"]

    def test_connect_remote_system_namespace_defaults_to_system(self) -> None:
        """Remote connections should default internal tables to __system."""
        db = connect(
            uri="db://test_db",
            api_key="test-key",
            host_override="https://phalanx.example.com",
        )

        assert db.system_namespace == ["__system"]

    def test_connect_remote_creates_checkpoint_store(self) -> None:
        """Test that checkpoint store is created for remote connections."""
        db = connect(
            uri="db://test_db",
            api_key="test-key",
            host_override="https://phalanx.example.com",
            upload_dir="test",
        )

        # Checkpoint store should be created
        assert db._checkpoint_store is not None

    def test_connect_local_vs_remote_differentiation(self, tmp_path: Path) -> None:
        """Test that local and remote URIs are handled differently."""
        # Local connection
        local_db = connect(tmp_path)
        assert not local_db.uri.startswith("db://")

        # Remote connection
        remote_db = connect(
            uri="db://remote_db",
            api_key="key",
            host_override="https://phalanx.example.com",
        )
        assert remote_db.uri == "db://remote_db"

    def test_connect_remote_flight_client_creation(self) -> None:
        """Test that flight_client is properly configured for remote connections."""
        db = connect(
            uri="db://test_db",
            api_key="test-key",
            host_override="https://phalanx.example.com:8080",
        )

        # Access flight_client to trigger creation
        with patch("flightsql.FlightSQLClient") as mock_flight:
            mock_flight.return_value = MagicMock()
            _ = db.flight_client

            # Verify FlightSQLClient was created with correct host
            mock_flight.assert_called_once()
            call_kwargs = mock_flight.call_args.kwargs
            assert call_kwargs.get("host") == "phalanx.example.com"


def test_connect_remote_does_not_probe_system_tables_for_upload_dir() -> None:
    """Remote connect should not touch system tables just to infer upload_dir."""
    with (
        patch("geneva.db.Uploader.get", side_effect=ValueError("unset")),
        patch("geneva.db.lancedb.connect") as mock_lancedb_connect,
    ):
        db = connect(
            uri="db://test_db",
            api_key="test-key",
            host_override="https://phalanx.example.com",
        )

    assert db.uri == "db://test_db"
    mock_lancedb_connect.assert_not_called()


def test_connect_namespace_does_not_probe_system_tables_for_upload_dir(
    tmp_path: Path,
) -> None:
    """Namespace connect should not create/open system tables just for upload_dir."""
    with (
        patch("geneva.db.Uploader.get", side_effect=ValueError("unset")),
        patch(
            "geneva.manifest.mgr.ManifestConfigManager",
            side_effect=AssertionError("should not probe manifest table"),
        ),
    ):
        db = connect(
            namespace_impl="dir",
            namespace_properties={"root": str(tmp_path)},
        )

    assert db.namespace_impl == "dir"


def test_ensure_system_namespace_exists_remote_defers_namespace_creation() -> None:
    with (
        patch("geneva.db._set_upload_dir"),
        patch("geneva.db.lancedb.connect") as mock_lancedb_connect,
    ):
        mock_remote = MagicMock()
        mock_lancedb_connect.return_value = mock_remote
        db = connect(
            uri="db://test_db",
            api_key="test-key",
            host_override="https://phalanx.example.com",
        )

        _ensure_system_namespace_exists(db)

    mock_lancedb_connect.assert_not_called()
    assert db._system_namespace_ensured is True


REMOTE_NAMESPACE_LOCATION_ERROR = (
    "Invalid input, Location must be provided when namespace is not empty"
)


def test_remote_system_tables_fallback_to_root_namespace_on_location_error() -> None:
    with (
        patch("geneva.db._set_upload_dir"),
        patch("geneva.utils.schema.alter_or_create_table") as mock_alter_or_create,
    ):
        db = connect(
            uri="db://test_db",
            api_key="test-key",
            host_override="https://phalanx.example.com",
        )
        db.migrate_legacy_system_table_if_needed = MagicMock()
        mock_table = MagicMock()
        mock_alter_or_create.side_effect = [
            RuntimeError(REMOTE_NAMESPACE_LOCATION_ERROR),
            mock_table,
        ]

        mgr = ManifestConfigManager(db)

    assert mgr.table is mock_table
    assert db.system_namespace == ["__system"]
    assert db.effective_system_namespace == []
    assert mock_alter_or_create.call_args_list[0].kwargs["namespace"] == ["__system"]
    assert mock_alter_or_create.call_args_list[1].kwargs["namespace"] == []


def test_remote_system_tables_fallback_when_probe_requires_location() -> None:
    with (
        patch("geneva.db._set_upload_dir"),
        patch("geneva.utils.schema.alter_or_create_table") as mock_alter_or_create,
    ):
        db = connect(
            uri="db://test_db",
            api_key="test-key",
            host_override="https://phalanx.example.com",
        )
        db.migrate_legacy_system_table_if_needed = MagicMock(
            side_effect=[
                RuntimeError(REMOTE_NAMESPACE_LOCATION_ERROR),
                None,
            ]
        )
        mock_table = MagicMock()
        mock_alter_or_create.return_value = mock_table

        mgr = ManifestConfigManager(db)

    assert mgr.table is mock_table
    assert db.system_namespace == ["__system"]
    assert db.effective_system_namespace == []
    assert db.migrate_legacy_system_table_if_needed.call_args_list[0].kwargs[
        "target_namespace"
    ] == ["__system"]
    assert (
        db.migrate_legacy_system_table_if_needed.call_args_list[1].kwargs[
            "target_namespace"
        ]
        == []
    )


def test_table_reference_open_db_preserves_remote_system_namespace() -> None:
    """Reopened remote refs should keep their explicit system namespace."""
    with patch("geneva.db._set_upload_dir"):
        ref = TableReference(
            table_id=["tbl"],
            version=None,
            db_uri="db://test_db",
            system_namespace=["prod", "system"],
        )

        db = ref.open_db()

    assert db.system_namespace == ["prod", "system"]


def test_table_reference_open_db_preserves_remote_credentials() -> None:
    with patch("geneva.table.connect") as mock_connect:
        ref = TableReference(
            table_id=["tbl"],
            version=None,
            db_uri="db://test_db",
            api_key="test-key",
            host_override="https://phalanx.example.com",
        )

        ref.open_db()

    assert mock_connect.call_args.kwargs["api_key"] == "test-key"
    assert mock_connect.call_args.kwargs["host_override"] == (
        "https://phalanx.example.com"
    )
    assert mock_connect.call_args.kwargs["initialize_upload_dir"] is False


def test_table_reference_open_db_does_not_probe_upload_dir(tmp_path: Path) -> None:
    with (
        patch("geneva.db.Uploader.get", side_effect=ValueError("unset")),
        patch(
            "geneva.manifest.mgr.ManifestConfigManager",
            side_effect=AssertionError("should not probe manifest table"),
        ),
    ):
        ref = TableReference(
            table_id=["tbl"],
            version=None,
            db_uri=str(tmp_path),
        )

        ref.open_db()


@pytest.mark.asyncio
async def test_table_reference_open_system_db_async_preserves_remote_credentials() -> (
    None
):
    with patch("geneva.table.connect_async") as mock_connect_async:
        ref = TableReference(
            table_id=["tbl"],
            version=None,
            db_uri="db://test_db",
            api_key="test-key",
            host_override="https://phalanx.example.com",
        )

        await ref.open_system_db_async()

    assert mock_connect_async.call_args.kwargs["api_key"] == "test-key"
    assert (
        mock_connect_async.call_args.kwargs["host_override"]
        == "https://phalanx.example.com"
    )


def test_table_reference_checkpoint_store_uses_remote_table_location() -> None:
    with (
        patch("geneva.table.resolve_table_physical_uri") as mock_resolve_uri,
        patch("geneva.checkpoint.LanceCheckpointStore") as mock_checkpoint_store,
    ):
        mock_resolve_uri.return_value = "s3://bucket/db/tbl.lance"

        ref = TableReference(
            table_id=["tbl"],
            version=None,
            db_uri="db://test_db",
            api_key="test-key",
            host_override="https://phalanx.example.com",
        )

        ref.open_checkpoint_store()

    assert mock_resolve_uri.call_args.kwargs["api_key"] == "test-key"
    assert (
        mock_resolve_uri.call_args.kwargs["host_override"]
        == "https://phalanx.example.com"
    )
    assert mock_checkpoint_store.call_args.args[0] == "s3://bucket/db/tbl.lance/_ckp"


def test_table_reference_open_db_preserves_explicit_empty_remote_system_namespace() -> (
    None
):
    """Explicit root namespace should not be rewritten to __system."""
    with patch("geneva.db._set_upload_dir"):
        ref = TableReference(
            table_id=["tbl"],
            version=None,
            db_uri="db://test_db",
            system_namespace=[],
        )

        db = ref.open_db()

    assert ref.system_table_namespace == []
    assert db.system_namespace == []


def test_colocated_system_table_ref_reuses_physical_db_root() -> None:
    with patch("geneva.table.connect") as mock_connect:
        ref = TableReference(
            table_id=["tbl"],
            version=None,
            db_uri="s3://bucket/db",
            api_key="test-key",
            host_override="http://localhost:10024",
            system_namespace=[],
            co_located_system_tables=True,
        )

        ref.as_system_table(GENEVA_JOBS_TABLE_NAME).open_db()

    assert mock_connect.call_args.args[0] == "s3://bucket/db"
    assert "api_key" not in mock_connect.call_args.kwargs
    assert "host_override" not in mock_connect.call_args.kwargs
    assert mock_connect.call_args.kwargs["is_system_db_root"] is False


class TestConnectConfig:
    """Tests for connect() configuration loading."""

    def test_connect_loads_config_defaults(self, tmp_path: Path) -> None:
        """Test that connect() loads defaults from config."""
        db = connect(tmp_path)
        # Default region should be set
        assert db._region == "us-east-1"

    def test_connect_explicit_params_override_config(self, tmp_path: Path) -> None:
        """Test that explicit parameters override config defaults."""
        db = connect(tmp_path, region="eu-west-1")
        assert db._region == "eu-west-1"


class TestUploadBucket:
    """Tests for upload_bucket configuration (Phalanx security model)."""

    def test_connect_with_upload_bucket(self, tmp_path: Path) -> None:
        """Test that upload_bucket is stored on Connection object."""
        db = connect(
            tmp_path,
            upload_dir="s3://my-upload-bucket/manifests",
        )
        assert db._upload_dir == "s3://my-upload-bucket/manifests"

    def test_connect_without_upload_bucket(self, tmp_path: Path) -> None:
        """Test that upload_bucket defaults to None."""
        db = connect(tmp_path)
        assert db._upload_dir is None

    def test_upload_bucket_with_remote_connection(self) -> None:
        """Test upload_bucket with remote db:// URI (Phalanx scenario)."""
        db = connect(
            uri="db://my_database",
            api_key="test-api-key",
            host_override="https://phalanx.example.com",
            upload_dir="gs://enterprise-upload-bucket/manifests",
        )
        assert db._upload_dir == "gs://enterprise-upload-bucket/manifests"
        assert db.uri == "db://my_database"

    def test_upload_bucket_serialization(self, tmp_path: Path) -> None:
        """Test that upload_bucket is properly serialized/deserialized."""
        db = connect(
            tmp_path,
            upload_dir="s3://my-upload-bucket/manifests",
        )

        # Get serialized state
        state = db.__getstate__()
        assert state["upload_dir"] == "s3://my-upload-bucket/manifests"

        # Create new connection from state
        new_db = connect(tmp_path)
        new_db.__setstate__(state)
        assert new_db._upload_dir == "s3://my-upload-bucket/manifests"

    def test_namespace_connection_serialization(self, tmp_path: Path) -> None:
        """Namespace connection state should survive a pickle-style round trip."""
        db = connect(
            namespace_impl="dir",
            namespace_properties={"root": str(tmp_path)},
            system_namespace=["prod", "system"],
        )

        state = db.__getstate__()
        assert state["namespace_impl"] == "dir"
        assert state["namespace_properties"] == {"root": str(tmp_path)}

        new_db = connect(tmp_path / "other")
        new_db.__setstate__(state)
        assert new_db.namespace_impl == "dir"
        assert new_db.namespace_properties == {"root": str(tmp_path)}
        assert new_db.system_namespace == ["prod", "system"]

    def test_define_manifest_uses_upload_bucket(self, tmp_path: Path) -> None:
        """Test that define_manifest uses upload_bucket when configured."""
        db = connect(
            tmp_path,
            upload_dir="s3://my-upload-bucket/manifests",
        )

        # Mock the Uploader to capture how it's instantiated
        with patch("geneva.db.Uploader") as mock_uploader_class:
            mock_uploader = MagicMock()
            mock_uploader_class.return_value = mock_uploader

            # Mock upload_local_env to avoid actual file operations
            with patch("geneva.db.upload_local_env") as mock_upload:
                mock_upload.return_value.__enter__ = MagicMock(return_value=[])
                mock_upload.return_value.__exit__ = MagicMock(return_value=None)

                # Mock the manifest manager
                with patch.object(db, "_manifest_manager") as mock_mgr:
                    mock_mgr.get_table.return_value = MagicMock()
                    db._manifest_manager = mock_mgr

                    # Create a minimal manifest
                    from geneva.manifest.mgr import GenevaManifest

                    manifest = GenevaManifest(name="test_manifest")

                    db.define_manifest("test_manifest", manifest)

                    # Verify Uploader was called with upload_dir (not table context)
                    mock_uploader_class.assert_called_once_with(
                        upload_dir="s3://my-upload-bucket/manifests"
                    )


@pytest.mark.parametrize(
    ("tbl_uri", "expected"),
    [
        # Simple bucket
        ("s3://bucket/table.lance", "s3://bucket"),
        # Bucket with path
        (
            "s3://bucket/path/table.lance",
            "s3://bucket/path",
        ),
        # local
        (
            "/foo/bar/table.lance",
            "/foo/bar",
        ),
        # GCS bucket
        ("gs://gcs-bucket/foo/table.lance", "gs://gcs-bucket/foo"),
        # Remote URI
        ("db://foo/table.lance", "db://foo"),
        # S3+DDB with query params
        (
            "s3+ddb://bucket/path/table.lance?ddbTableName=external-manifest",
            "s3+ddb://bucket/path?ddbTableName=external-manifest",
        ),
        # S3+DDB nested path with query params
        (
            "s3+ddb://bucket/path/to/table.lance?ddbTableName=my-table",
            "s3+ddb://bucket/path/to?ddbTableName=my-table",
        ),
    ],
)
def test_get_db_uri_from_table(tbl_uri, expected) -> None:
    from geneva.db import _get_db_uri

    result = _get_db_uri(tbl_uri)
    assert result == expected


@pytest.mark.parametrize(
    ("db_uri", "table_id", "expected_suffix"),
    [
        # Simple S3 URI
        (
            "s3://bucket/path",
            ["my_table"],
            "s3://bucket/path/my_table.lance/_geneva_uploads",
        ),
        # S3+DDB URI with query params — query string must come after the full path
        (
            "s3+ddb://bucket/path?ddbTableName=external-manifest",
            ["my_table"],
            "s3+ddb://bucket/path/my_table.lance/_geneva_uploads?ddbTableName=external-manifest",
        ),
    ],
)
def test_uploader_db_uri_with_query_params(db_uri, table_id, expected_suffix) -> None:
    from geneva.packager.uploader import Uploader

    uploader = Uploader(db_uri=db_uri, table_id=table_id)
    assert uploader.upload_dir == expected_suffix



# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors

"""Frame extraction UDF for blob column testing in materialized views."""

from geneva.udfs.video.frame_extractor import ExtractFirstFrame

__all__ = ["ExtractFirstFrame"]



# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors

"""Pipeline helpers for WavLM TBR ONNX speaker-embedding workflows."""

from __future__ import annotations

import argparse
import contextlib
from contextlib import AbstractContextManager
from typing import TYPE_CHECKING, Any

import numpy as np

import geneva
from e2e.onnx_audio._pipeline_utils import _default_local_context, _open_or_create_table
from geneva.udfs.audio.wavlm_tbr_onnx import wavlm_tbr_embedding_udf

if TYPE_CHECKING:
    from geneva.transformer import UDF

DEFAULT_DATASET = "openslr/librispeech_asr"
DEFAULT_CONFIG = "clean"
DEFAULT_SPLIT = "test"
DEFAULT_NUM_SPEAKERS = 4
DEFAULT_SAMPLES_PER_SPEAKER = 100
DEFAULT_SEED = 13
DEFAULT_SAMPLE_RATE = 16000
DEFAULT_TABLE = "librispeech_speaker_embeddings"
DEFAULT_AUDIO_COLUMN = "audio"
DEFAULT_SR_COLUMN = "sampling_rate"
DEFAULT_EMBED_COLUMN = "embedding"
DEFAULT_MANIFEST_NAME = "speaker-wavlm-tbr-onnx-udf"


def ensure_manifest(conn: geneva.Connection, manifest_name: str) -> None:
    """Define a manifest suitable for running the WavLM UDF on remote workers."""
    from geneva.manifest.builder import GenevaManifestBuilder

    manifest = (
        GenevaManifestBuilder.create(manifest_name)
        .pip(
            [
                "numpy>=1.26",
                "onnxruntime>=1.17,<1.24",
            ]
        )
        .build()
    )
    conn.define_manifest(manifest_name, manifest)


def _load_dataset(
    dataset: str,
    config: str,
    split: str,
    cache_dir: str | None,
) -> Any:
    try:
        from datasets import load_dataset
    except ImportError as exc:
        raise ImportError(
            "The `datasets` library is required. Install via `pip install datasets`."
        ) from exc

    ds = load_dataset(dataset, config, split=split, cache_dir=cache_dir, streaming=True)
    return ds.decode(False)


def _sample_rows_single_pass(
    dataset: Any,
    *,
    samples_per_speaker: int,
    seed: int,
    num_speakers: int,
) -> list[dict[str, Any]]:
    """Single-pass reservoir sampling over a streaming LibriSpeech iterator."""

    sample_rng = np.random.default_rng(seed + 1)
    speaker_counts: dict[int, int] = {}
    reservoirs: dict[int, list[dict[str, Any]]] = {}

    for row in dataset:
        raw_speaker = row.get("speaker_id")
        if raw_speaker is None:
            continue
        speaker_id = int(raw_speaker)
        speaker_counts[speaker_id] = speaker_counts.get(speaker_id, 0) + 1

        reservoir = reservoirs.setdefault(speaker_id, [])
        if len(reservoir) < samples_per_speaker:
            reservoir.append(row)
            continue

        choice = sample_rng.integers(0, speaker_counts[speaker_id])
        if choice < samples_per_speaker:
            reservoir[choice] = row

    eligible = [
        speaker_id
        for speaker_id, count in speaker_counts.items()
        if count >= samples_per_speaker
    ]
    if len(eligible) < num_speakers:
        raise ValueError(
            f"Only {len(eligible)} speakers have >= {samples_per_speaker} samples; "
            f"requested {num_speakers}."
        )

    select_rng = np.random.default_rng(seed)
    selected_speakers = select_rng.choice(eligible, size=num_speakers, replace=False)

    rows: list[dict[str, Any]] = []
    for raw_speaker_id in selected_speakers.tolist():
        speaker_id = int(raw_speaker_id)
        samples = reservoirs.get(speaker_id, [])
        if len(samples) != samples_per_speaker:
            raise ValueError(
                "Not enough samples for selected speaker "
                f"{speaker_id}: {len(samples)}"
            )
        rows.extend(samples)

    sample_rng.shuffle(rows)
    return rows


def _decode_audio(record: dict[str, Any], sample_rate: int) -> tuple[np.ndarray, int]:
    try:
        import soundfile as sf
    except ImportError as exc:
        raise ImportError(
            "soundfile is required for audio decoding; "
            "install via `pip install soundfile`."
        ) from exc

    audio = record.get("audio")
    if audio is None:
        raise ValueError("Missing audio data in record.")

    data = None
    sr = None
    if isinstance(audio, dict):
        audio_bytes = audio.get("bytes")
        audio_path = audio.get("path")
        if audio_bytes:
            import io

            with sf.SoundFile(io.BytesIO(audio_bytes)) as file:
                sr = file.samplerate
                data = file.read(dtype="float32")
        elif audio_path:
            with sf.SoundFile(audio_path) as file:
                sr = file.samplerate
                data = file.read(dtype="float32")
    elif isinstance(audio, str):
        with sf.SoundFile(audio) as file:
            sr = file.samplerate
            data = file.read(dtype="float32")

    if data is None or sr is None:
        raise ValueError("Unable to decode audio; missing path or bytes.")

    if data.ndim == 2:
        data = data.mean(axis=0)

    if sr != sample_rate:
        raise ValueError(
            f"Expected {sample_rate} Hz audio; decoded {sr} Hz. "
            "Resample the dataset to 16 kHz before running."
        )

    return data.astype(np.float32), int(sr)


def _build_rows(
    records: list[dict[str, Any]],
    sample_rate: int,
    audio_column: str,
    sample_rate_column: str,
) -> list[dict[str, Any]]:
    rows: list[dict[str, Any]] = []
    for idx, row in enumerate(records):
        audio_array, decoded_sr = _decode_audio(row, sample_rate)
        rows.append(
            {
                "row_id": idx,
                "id": row.get("id"),
                "speaker_id": row.get("speaker_id"),
                "chapter_id": row.get("chapter_id"),
                "file": row.get("file"),
                "text": row.get("text"),
                audio_column: audio_array.tolist(),
                sample_rate_column: decoded_sr,
            }
        )
    return rows


def build_librispeech_rows(
    *,
    dataset: str = DEFAULT_DATASET,
    config: str = DEFAULT_CONFIG,
    split: str = DEFAULT_SPLIT,
    cache_dir: str | None = None,
    seed: int = DEFAULT_SEED,
    num_speakers: int = DEFAULT_NUM_SPEAKERS,
    samples_per_speaker: int = DEFAULT_SAMPLES_PER_SPEAKER,
    sample_rate: int = DEFAULT_SAMPLE_RATE,
    audio_column: str = DEFAULT_AUDIO_COLUMN,
    sample_rate_column: str = DEFAULT_SR_COLUMN,
) -> list[dict[str, Any]]:
    """Sample LibriSpeech records and return table rows."""
    dataset_iter = _load_dataset(dataset, config, split, cache_dir)
    sampled_records = _sample_rows_single_pass(
        dataset_iter,
        samples_per_speaker=samples_per_speaker,
        seed=seed,
        num_speakers=num_speakers,
    )
    return _build_rows(
        sampled_records,
        sample_rate,
        audio_column,
        sample_rate_column,
    )


def ensure_table(
    conn: geneva.Connection,
    table_name: str,
    rows: list[dict[str, Any]],
) -> geneva.Table:
    """Create or open a table for speaker-embedding runs."""
    return _open_or_create_table(conn, table_name, rows)


def run_wavlm_tbr_pipeline(
    table: geneva.Table,
    *,
    model_dir: str | None = None,
    embedding_udf: UDF | None = None,
    audio_column: str = DEFAULT_AUDIO_COLUMN,
    sample_rate_column: str = DEFAULT_SR_COLUMN,
    embedding_column: str = DEFAULT_EMBED_COLUMN,
    sample_rate: int = DEFAULT_SAMPLE_RATE,
    max_seconds: float = 30.0,
    num_gpus: float = 0.0,
    checkpoint_size: int = 8,
    concurrency: int = 4,
    rebind_udf: bool = True,
    create_index: bool = True,
    index_type: str = "IVF_PQ",
    metric: str = "cosine",
    num_partitions: int | None = None,
    num_sub_vectors: int | None = 128,
    replace_index: bool = True,
    context: AbstractContextManager[None] | None = None,
) -> geneva.Table:
    """
    Add/update a WavLM embedding column, backfill it, and optionally index it.

    Pass either ``model_dir`` (to build the ONNX UDF) or ``embedding_udf`` (for
    testing/custom UDF implementations).
    """
    if embedding_udf is None:
        if model_dir is None:
            raise ValueError("Either model_dir or embedding_udf must be provided")
        embedding_udf = wavlm_tbr_embedding_udf(
            model_dir=model_dir,
            audio_column=audio_column,
            sample_rate=sample_rate,
            sample_rate_column=sample_rate_column,
            max_seconds=max_seconds,
            num_gpus=num_gpus,
        )

    if embedding_column not in table.schema.names:
        table.add_columns({embedding_column: embedding_udf})
    elif rebind_udf:
        table.alter_columns({"path": embedding_column, "udf": embedding_udf})

    active_context = context or contextlib.nullcontext()
    with active_context:
        table.backfill(
            embedding_column,
            checkpoint_size=checkpoint_size,
            concurrency=concurrency,
        )

    if create_index:
        index_kwargs: dict[str, Any] = {
            "metric": metric,
            "vector_column_name": embedding_column,
            "replace": replace_index,
            "index_type": index_type,
        }
        if num_partitions is not None:
            index_kwargs["num_partitions"] = num_partitions
        if num_sub_vectors is not None and "PQ" in index_type:
            index_kwargs["num_sub_vectors"] = num_sub_vectors
        table.create_index(**index_kwargs)

    return table


def main() -> None:
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("--db-uri", default="./db", help="Geneva database URI")
    parser.add_argument("--table", default=DEFAULT_TABLE, help="Table name")
    parser.add_argument(
        "--model-dir",
        required=True,
        help="Directory containing model.onnx",
    )
    parser.add_argument(
        "--dataset",
        default=DEFAULT_DATASET,
        help="Hugging Face dataset name",
    )
    parser.add_argument(
        "--dataset-cache-dir",
        default=None,
        help="Optional Hugging Face cache directory",
    )
    parser.add_argument(
        "--seed",
        type=int,
        default=DEFAULT_SEED,
        help="Random seed for speaker sampling",
    )
    parser.add_argument(
        "--num-speakers",
        type=int,
        default=DEFAULT_NUM_SPEAKERS,
        help="Number of speakers to sample",
    )
    parser.add_argument(
        "--samples-per-speaker",
        type=int,
        default=DEFAULT_SAMPLES_PER_SPEAKER,
        help="Number of utterances per speaker",
    )
    parser.add_argument(
        "--sample-rate",
        type=int,
        default=DEFAULT_SAMPLE_RATE,
        help="Target sample rate for audio",
    )
    parser.add_argument(
        "--audio-column",
        default=DEFAULT_AUDIO_COLUMN,
        help="Audio column name",
    )
    parser.add_argument(
        "--sample-rate-column",
        default=DEFAULT_SR_COLUMN,
        help="Sample rate column name",
    )
    parser.add_argument(
        "--embedding-column",
        default=DEFAULT_EMBED_COLUMN,
        help="Embedding output column name",
    )
    parser.add_argument(
        "--max-seconds",
        type=float,
        default=30.0,
        help="Truncate audio longer than this many seconds",
    )
    parser.add_argument("--num-gpus", type=float, default=0.0, help="GPU fraction")
    parser.add_argument(
        "--checkpoint-size",
        "--batch-size",
        action="store",
        dest="checkpoint_size",
        type=int,
        default=8,
        help="Backfill checkpoint size",
    )
    parser.add_argument(
        "--concurrency",
        type=int,
        default=4,
        help="Backfill concurrency",
    )
    parser.add_argument(
        "--cluster",
        default=None,
        help="Optional cluster name for remote execution",
    )
    parser.add_argument(
        "--manifest-name",
        default=None,
        help="Manifest name to use when running on a cluster",
    )
    parser.add_argument(
        "--define-manifest",
        action="store_true",
        help="Define/update the manifest before running",
    )
    parser.add_argument(
        "--no-create-table",
        action="store_true",
        help="Do not create a dataset-backed table when missing",
    )
    parser.add_argument(
        "--no-rebind-udf",
        action="store_true",
        help="Do not alter an existing embedding column to the latest UDF",
    )
    parser.add_argument(
        "--skip-index",
        action="store_true",
        help="Skip creating a vector index",
    )
    parser.add_argument(
        "--index-type",
        default="IVF_PQ",
        choices=["IVF_FLAT", "IVF_PQ", "IVF_HNSW_SQ", "IVF_HNSW_PQ"],
        help="Vector index type",
    )
    parser.add_argument(
        "--metric",
        default="cosine",
        help="Vector index distance metric",
    )
    parser.add_argument(
        "--num-partitions",
        type=int,
        default=None,
        help="Number of IVF partitions",
    )
    parser.add_argument(
        "--num-sub-vectors",
        type=int,
        default=128,
        help="Number of PQ sub-vectors",
    )
    parser.add_argument(
        "--no-replace-index",
        action="store_true",
        help="Do not replace an existing vector index",
    )
    args = parser.parse_args()

    conn = geneva.connect(args.db_uri)

    if args.no_create_table:
        table = conn.open_table(args.table)
    else:
        rows = build_librispeech_rows(
            dataset=args.dataset,
            config=DEFAULT_CONFIG,
            split=DEFAULT_SPLIT,
            cache_dir=args.dataset_cache_dir,
            seed=args.seed,
            num_speakers=args.num_speakers,
            samples_per_speaker=args.samples_per_speaker,
            sample_rate=args.sample_rate,
            audio_column=args.audio_column,
            sample_rate_column=args.sample_rate_column,
        )
        table = ensure_table(conn, args.table, rows)

    context: AbstractContextManager[None]
    if args.cluster:
        manifest_name = args.manifest_name or DEFAULT_MANIFEST_NAME
        if args.define_manifest or args.manifest_name is None:
            ensure_manifest(conn, manifest_name)
        context = conn.context(cluster=args.cluster, manifest=manifest_name)
    else:
        context = _default_local_context(conn)

    run_wavlm_tbr_pipeline(
        table,
        model_dir=args.model_dir,
        audio_column=args.audio_column,
        sample_rate_column=args.sample_rate_column,
        embedding_column=args.embedding_column,
        sample_rate=args.sample_rate,
        max_seconds=args.max_seconds,
        num_gpus=args.num_gpus,
        checkpoint_size=args.checkpoint_size,
        concurrency=args.concurrency,
        rebind_udf=not args.no_rebind_udf,
        create_index=not args.skip_index,
        index_type=args.index_type,
        metric=args.metric,
        num_partitions=args.num_partitions,
        num_sub_vectors=args.num_sub_vectors,
        replace_index=not args.no_replace_index,
        context=context,
    )


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors

from .builder import (
    CondaManifestBuilder,
    PipManifestBuilder,
    SiteManifestBuilder,
)
from .mgr import GenevaManifest, ManifestConfigManager

__all__ = [
    "CondaManifestBuilder",
    "GenevaManifest",
    "ManifestConfigManager",
    "PipManifestBuilder",
    "SiteManifestBuilder",
]



# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors
import contextlib
import copy
import logging
import threading
from collections.abc import Callable, Iterable
from datetime import datetime, timedelta, timezone
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, Any, TypeVar
from urllib.parse import urlparse, urlunparse

import attrs
import lancedb
import pyarrow as pa
from lance_namespace import LanceNamespace
from lance_namespace import connect as namespace_connect
from lancedb import DBConnection, LanceNamespaceDBConnection
from lancedb.common import DATA, Credential
from lancedb.pydantic import LanceModel
from lancedb.table import Table as LanceDBTable
from lancedb.util import get_uri_scheme
from overrides import override
from yarl import URL

from geneva import DEFAULT_UPLOAD_DIR
from geneva.checkpoint import CheckpointStore
from geneva.cluster import GenevaClusterType
from geneva.config import ConfigBase
from geneva.packager import DockerUDFPackager, UDFPackager
from geneva.packager.autodetect import upload_local_env
from geneva.packager.uploader import Uploader

if TYPE_CHECKING:
    import lance

    from geneva.cluster.mgr import ClusterConfigManager, GenevaCluster
    from geneva.jobs.jobs import JobStateManager
    from geneva.manifest.mgr import GenevaManifest, ManifestConfigManager
    from geneva.query import GenevaQueryBuilder
    from geneva.table import Table
    from geneva.transformer import UDTF, ScalarUDTF

_LOG = logging.getLogger(__name__)

T = TypeVar("T")

SYSTEM_NAMESPACE = "__system"
SYSTEM_TABLE_MIGRATIONS_TABLE_NAME = "geneva_migrations"
SYSTEM_TABLE_MIGRATING = "migrating"
SYSTEM_TABLE_COMPLETED = "completed"


@attrs.define
class SystemTableMigrationRecord:
    table_name: str = attrs.field()
    status: str = attrs.field()
    updated_at: datetime = attrs.field(
        factory=lambda: datetime.now(timezone.utc),
        metadata={"pa_type": pa.timestamp("us", tz="UTC")},
    )


def uses_namespace_for_system_tables(
    uri: str | None,
    namespace_impl: str | None,
) -> bool:
    """Whether a connection/reference should locate system tables via namespaces."""
    return namespace_impl is not None or bool(uri and str(uri).startswith("db://"))


def infer_system_namespace(
    *,
    uri: str | None,
    namespace_impl: str | None,
    system_namespace: list[str] | None,
) -> list[str]:
    if system_namespace is not None:
        return system_namespace
    if uses_namespace_for_system_tables(uri, namespace_impl):
        return [SYSTEM_NAMESPACE]
    return []


def infer_system_db_uri(
    *,
    uri: str | None,
    namespace_impl: str | None,
) -> str | None:
    if uri is None:
        return None
    if uses_namespace_for_system_tables(uri, namespace_impl):
        return uri
    return str(URL(str(uri)) / SYSTEM_NAMESPACE)


# Lance internal columns that are always available but not in the user schema
_LANCE_INTERNAL_COLUMNS = {"_rowid", "_rowaddr"}


def has_stable_row_ids(fragments: "Iterable[lance.LanceFragment]") -> bool:
    """Check if Lance fragments have stable row IDs enabled.

    Stable row IDs are indicated by presence of row_id_meta on fragment metadata.
    This is a Lance feature (added in v0.21.0) that ensures row identifiers remain
    constant even when table operations like compaction reorganize the physical data.

    Parameters
    ----------
    fragments : Iterable[lance.LanceFragment]
        Lance fragments to check (from dataset.get_fragments())

    Returns
    -------
    bool
        True if any fragment has stable row IDs enabled, False otherwise
    """
    return any(frag.metadata.row_id_meta is not None for frag in fragments)


class Connection(DBConnection):
    """Geneva Connection."""

    def __init__(
        self,
        uri: str,
        *,
        region: str = "us-east-1",
        api_key: Credential | None = None,
        host_override: str | None = None,
        storage_options: dict[str, str] | None = None,
        checkpoint_store: CheckpointStore | None = None,
        packager: UDFPackager | None = None,
        namespace_impl: str | None = None,
        namespace_properties: dict[str, str] | None = None,
        system_namespace: list[str],
        is_system_db_root: bool = False,
        upload_dir: str | None = None,
        **kwargs,
    ) -> None:
        super().__init__()

        self._uri = uri
        self._region = region
        self._api_key = api_key
        self._host_override = host_override
        self._storage_options = storage_options
        self._ldb: DBConnection | None = None
        self._checkpoint_store = checkpoint_store
        self._packager = packager or DockerUDFPackager()
        self.namespace_impl = namespace_impl
        self.namespace_properties = namespace_properties
        self._is_system_db_root = is_system_db_root
        # Use namespace-based system tables for namespace and remote connections.
        # For local/object-store direct connections, system tables live under a
        # sibling "__system" database URI.
        uses_namespace_system_tables = uses_namespace_for_system_tables(
            uri, namespace_impl
        )
        self.system_namespace = system_namespace if uses_namespace_system_tables else []
        self._upload_dir = upload_dir

        self._jobs_manager: JobStateManager | None = None
        self._cluster_manager: ClusterConfigManager | None = None
        self._manifest_manager: ManifestConfigManager | None = None
        self._flight_client: Any | None = None
        self._namespace_connection: LanceNamespaceDBConnection | None = None
        self._system_table_conn: Connection | None = None
        self._system_table_conn_lock = threading.Lock()
        self._system_table_migrations: set[tuple[str, tuple[str, ...], str]] = set()
        self._system_namespace_ensured = False
        self._system_namespace_fallback: list[str] | None = None
        self._kwargs = kwargs

    def __repr__(self) -> str:
        return f"<Geneva uri={self.uri}>"

    def __getstate__(self) -> dict:
        return {
            "uri": self._uri,
            "api_key": self._api_key,
            "host_override": self._host_override,
            "storage_options": self._storage_options,
            "region": self._region,
            "namespace_impl": self.namespace_impl,
            "namespace_properties": self.namespace_properties,
            "system_namespace": self.effective_system_namespace,
            "is_system_db_root": self._is_system_db_root,
            "upload_dir": self._upload_dir,
        }

    def __setstate__(self, state) -> None:
        self.__init__(state.pop("uri"), **state)

    def __enter__(self) -> "Connection":
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.close()
        return None  # Don't suppress exceptions

    def close(self) -> None:
        """Close the connection."""
        if self._flight_client is not None:
            self._flight_client.close()
        if self._ldb is not None and hasattr(self._ldb, "_conn"):
            # go to the async client and eagerly close the connection
            self._ldb._conn.close()  # type: ignore[attr-defined]
        if self._system_table_conn is not None:
            self._system_table_conn.close()

    def _uses_namespace_for_system_tables(self) -> bool:
        """Whether system tables should use namespace APIs on this connection."""
        return uses_namespace_for_system_tables(self._uri, self.namespace_impl)

    def _is_system_table_db_root(self) -> bool:
        """Whether this direct connection already points at the system DB root."""
        return self._is_system_db_root

    @property
    def user_db_uri(self) -> str:
        """Original user database URI for this connection."""
        if not self._is_system_table_db_root():
            return str(self.uri)

        parsed = urlparse(str(self.uri))
        path = parsed.path.rstrip("/")
        if path.endswith(f"/{SYSTEM_NAMESPACE}"):
            path = path[: -(len(SYSTEM_NAMESPACE) + 1)] or "/"
        return urlunparse(parsed._replace(path=path))

    @property
    def effective_system_namespace(self) -> list[str]:
        """Effective namespace after backend-compatibility fallback."""
        if self._system_namespace_fallback is not None:
            return self._system_namespace_fallback
        return self.system_namespace

    @property
    def system_table_db_uri(self) -> str:
        """Database URI where system tables should live."""
        if self._uses_namespace_for_system_tables() or self._is_system_table_db_root():
            return str(self.uri)
        return str(URL(str(self.uri)) / SYSTEM_NAMESPACE)

    @property
    def system_table_connection(self) -> "Connection":
        """Connection used to access system tables."""
        if self._uses_namespace_for_system_tables() or self._is_system_table_db_root():
            return self

        if self._system_table_conn is None:
            with self._system_table_conn_lock:
                if self._system_table_conn is None:
                    self._system_table_conn = Connection(
                        self.system_table_db_uri,
                        region=self._region,
                        api_key=self._api_key,
                        host_override=self._host_override,
                        storage_options=self._storage_options,
                        checkpoint_store=self._checkpoint_store,
                        packager=self._packager,
                        namespace_impl=None,
                        namespace_properties=None,
                        system_namespace=[],
                        is_system_db_root=True,
                        upload_dir=self._upload_dir,
                        **self._kwargs,
                    )
        return self._system_table_conn

    def resolve_system_table_location(
        self, namespace: list[str] | None = None
    ) -> tuple["Connection", list[str]]:
        """Resolve the target connection/namespace for system table operations."""
        if namespace is not None:
            return self, namespace
        if self._uses_namespace_for_system_tables():
            return self, self.effective_system_namespace
        return self.system_table_connection, []

    def _run_system_table_operation(
        self,
        table_name: str,
        operation: Callable[["Connection", list[str]], T],
        *,
        namespace: list[str] | None = None,
        migrate_legacy: bool = True,
    ) -> T:
        """Run a system-table operation with one remote fallback path."""

        def run_on_target(
            target_db: "Connection",
            target_namespace: list[str],
        ) -> T:
            if target_namespace and target_db._uses_namespace_for_system_tables():
                _ensure_system_namespace_exists(target_db)
            if migrate_legacy:
                self.migrate_legacy_system_table_if_needed(
                    table_name=table_name,
                    target_conn=target_db,
                    target_namespace=target_namespace,
                )
            return operation(target_db, target_namespace)

        target_db, target_namespace = self.resolve_system_table_location(namespace)
        try:
            return run_on_target(target_db, target_namespace)
        except Exception as e:
            if (
                namespace is not None
                or not self.is_remote()
                or not target_namespace
                or not self._is_remote_namespace_location_error(e)
            ):
                raise

            _LOG.warning(
                "Remote backend rejected namespace-backed system table '%s'; "
                "falling back to root namespace for compatibility",
                table_name,
            )
            self._system_namespace_fallback = []
            fallback_db, fallback_namespace = self.resolve_system_table_location(None)
            return run_on_target(fallback_db, fallback_namespace)

    @staticmethod
    def _is_remote_namespace_location_error(exc: Exception) -> bool:
        return (
            "location must be provided when namespace is not empty" in str(exc).lower()
        )

    @staticmethod
    def _is_not_found_error(exc: Exception) -> bool:
        msg = str(exc).lower()
        return "not found" in msg or "does not exist" in msg

    @staticmethod
    def _table_exists(
        conn: "Connection", table_name: str, namespace: list[str] | None
    ) -> bool:
        ns = namespace if namespace is not None else []
        try:
            conn.open_table(table_name, namespace=ns)
            return True
        except Exception as e:
            if Connection._is_not_found_error(e):
                return False
            _LOG.debug(
                "Error checking existence of table '%s' in namespace=%s on uri=%s",
                table_name,
                ns,
                conn.uri,
                exc_info=True,
            )
            raise

    @staticmethod
    def _replace_table_from_source(
        *,
        src_conn: "Connection",
        src_namespace: list[str] | None,
        dst_conn: "Connection",
        dst_namespace: list[str] | None,
        table_name: str,
    ) -> None:
        src_ns = src_namespace if src_namespace is not None else []
        dst_ns = dst_namespace if dst_namespace is not None else []

        src_tbl = src_conn.open_table(table_name, namespace=src_ns)
        src_data = src_tbl.search().to_arrow()
        src_row_count = src_tbl.count_rows()

        if Connection._table_exists(dst_conn, table_name, dst_ns):
            dst_conn.drop_table(table_name, namespace=dst_ns)

        dst_conn.create_table(
            table_name,
            schema=src_tbl.schema,
            mode="overwrite",
            namespace=dst_ns,
        )

        if src_row_count > 0:
            dst_tbl = dst_conn.open_table(table_name, namespace=dst_ns)
            dst_tbl.add(src_data)

        dst_tbl = dst_conn.open_table(table_name, namespace=dst_ns)
        if dst_tbl.count_rows() != src_row_count:
            raise RuntimeError(
                f"Migration validation failed for '{table_name}': "
                f"expected {src_row_count} rows but copied {dst_tbl.count_rows()}"
            )

    def _legacy_system_table_candidates(
        self,
        target_conn: "Connection",
        target_namespace: list[str] | None,
    ) -> list[tuple["Connection", list[str]]]:
        dst_ns = target_namespace if target_namespace is not None else []
        candidates: list[tuple[Connection, list[str]]] = []

        if self._uses_namespace_for_system_tables():
            if self.namespace_impl is not None and dst_ns != ["default"]:
                candidates.append((self, ["default"]))
            if self.is_remote() and dst_ns != []:
                candidates.append((self, []))
        else:
            candidates.append((self, []))

        return [
            (legacy_conn, legacy_ns)
            for legacy_conn, legacy_ns in candidates
            if not (legacy_conn is target_conn and legacy_ns == dst_ns)
        ]

    def _find_legacy_system_table_source(
        self,
        *,
        table_name: str,
        target_conn: "Connection",
        target_namespace: list[str] | None,
    ) -> tuple["Connection", list[str]] | None:
        for legacy_conn, legacy_ns in self._legacy_system_table_candidates(
            target_conn,
            target_namespace,
        ):
            if self._table_exists(legacy_conn, table_name, legacy_ns):
                return legacy_conn, legacy_ns
        return None

    def _open_system_table_migrations_table(
        self, create_if_missing: bool
    ) -> tuple["Connection", LanceDBTable, list[str]] | None:
        from geneva.utils.schema import alter_or_create_table

        def open_or_create(
            target_db: "Connection",
            target_namespace: list[str],
        ) -> tuple["Connection", LanceDBTable, list[str]] | None:
            if not create_if_missing and not self._table_exists(
                target_db,
                SYSTEM_TABLE_MIGRATIONS_TABLE_NAME,
                target_namespace,
            ):
                return None

            table = alter_or_create_table(
                target_db,
                SYSTEM_TABLE_MIGRATIONS_TABLE_NAME,
                SystemTableMigrationRecord(
                    table_name="dummy",
                    status=SYSTEM_TABLE_COMPLETED,
                ),
                namespace=target_namespace,
            )
            return target_db, table, target_namespace

        return self._run_system_table_operation(
            SYSTEM_TABLE_MIGRATIONS_TABLE_NAME,
            open_or_create,
            migrate_legacy=False,
        )

    def _get_system_table_migration_status(self, table_name: str) -> str | None:
        table_info = self._open_system_table_migrations_table(create_if_missing=False)
        if table_info is None:
            return None
        _, table, _ = table_info
        rows = (
            table._ltbl.search()  # pyright: ignore[reportAttributeAccessIssue]
            .where(f"table_name = '{table_name}'")
            .limit(1)
            .to_arrow()
            .to_pylist()
        )
        if not rows:
            return None
        return rows[0]["status"]

    def _set_system_table_migration_status(self, table_name: str, status: str) -> None:
        table_info = self._open_system_table_migrations_table(create_if_missing=True)
        assert table_info is not None, (
            "migration table must exist when create_if_missing=True"
        )
        _, table, _ = table_info
        ltbl = table._ltbl  # pyright: ignore[reportAttributeAccessIssue]
        ltbl.delete(f"table_name = '{table_name}'")
        ltbl.add(
            [
                attrs.asdict(
                    SystemTableMigrationRecord(
                        table_name=table_name,
                        status=status,
                    )
                )
            ]
        )

    def _clear_system_table_migration_status(self, table_name: str) -> None:
        table_info = self._open_system_table_migrations_table(create_if_missing=False)
        if table_info is None:
            return
        _, table, _ = table_info
        ltbl = table._ltbl  # pyright: ignore[reportAttributeAccessIssue]
        ltbl.delete(f"table_name = '{table_name}'")

    def migrate_legacy_system_table_if_needed(
        self,
        *,
        table_name: str,
        target_conn: "Connection",
        target_namespace: list[str] | None,
    ) -> None:
        """Best-effort migration from legacy system-table locations.

        Legacy locations:
        - Local/object-store direct connections: <db>/<table>.lance (root)
        - Remote connections: root namespace []
        - Namespace connections: default namespace ["default"]
        """
        dst_ns = target_namespace if target_namespace is not None else []
        key = (table_name, tuple(dst_ns), str(target_conn.uri))
        if key in self._system_table_migrations:
            return
        if table_name == SYSTEM_TABLE_MIGRATIONS_TABLE_NAME:
            self._system_table_migrations.add(key)
            return

        status = self._get_system_table_migration_status(table_name)
        legacy_source = self._find_legacy_system_table_source(
            table_name=table_name,
            target_conn=target_conn,
            target_namespace=target_namespace,
        )

        if status == SYSTEM_TABLE_COMPLETED:
            self._system_table_migrations.add(key)
            return

        if status is None and legacy_source is None:
            self._system_table_migrations.add(key)
            return

        if legacy_source is None:
            if status == SYSTEM_TABLE_MIGRATING:
                _LOG.warning(
                    "Clearing stale migration marker for system table '%s' on uri=%s "
                    "namespace=%s because no legacy source table was found",
                    table_name,
                    target_conn.uri,
                    dst_ns,
                )
                self._clear_system_table_migration_status(table_name)
                self._system_table_migrations.add(key)
                return
            raise RuntimeError(
                f"System table '{table_name}' is marked as migrating but no legacy "
                "source table was found"
            )

        legacy_conn, legacy_ns = legacy_source
        self._set_system_table_migration_status(table_name, SYSTEM_TABLE_MIGRATING)
        _LOG.info(
            "Migrating legacy system table '%s' from namespace=%s uri=%s "
            "to namespace=%s uri=%s",
            table_name,
            legacy_ns,
            legacy_conn.uri,
            dst_ns,
            target_conn.uri,
        )
        self._replace_table_from_source(
            src_conn=legacy_conn,
            src_namespace=legacy_ns,
            dst_conn=target_conn,
            dst_namespace=dst_ns,
            table_name=table_name,
        )
        self._set_system_table_migration_status(table_name, SYSTEM_TABLE_COMPLETED)
        _LOG.warning(
            "Legacy system table '%s' remains at namespace=%s uri=%s after "
            "copying to uri=%s for rollback safety",
            table_name,
            legacy_ns,
            legacy_conn.uri,
            target_conn.uri,
        )
        self._system_table_migrations.add(key)

    def prepare_system_table_target(
        self,
        table_name: str,
        namespace: list[str] | None = None,
    ) -> tuple["Connection", list[str]]:
        """Resolve and prepare the destination for a system-table operation."""
        return self._run_system_table_operation(
            table_name,
            lambda target_db, target_namespace: (target_db, target_namespace),
            namespace=namespace,
        )

    def alter_or_create_system_table(
        self,
        table_name: str,
        model: Any,
        namespace: list[str] | None = None,
    ) -> tuple["Connection", LanceDBTable]:
        """Open or create a Geneva internal table in its resolved location."""
        from geneva.utils.schema import alter_or_create_table

        return self._run_system_table_operation(
            table_name,
            lambda target_db, target_namespace: (
                target_db,
                alter_or_create_table(
                    target_db,
                    table_name,
                    model,
                    namespace=target_namespace,
                ),
            ),
            namespace=namespace,
        )

    @cached_property
    def namespace_client(self) -> LanceNamespace | None:
        """Returns namespace client if using namespace connection."""
        if self.namespace_impl is not None and self.namespace_properties is not None:
            return namespace_connect(self.namespace_impl, self.namespace_properties)
        return None

    @cached_property
    def _connect(self) -> DBConnection:
        conn = self._get_conn(uri=self.uri)

        if self.namespace_impl is not None:
            if self._namespace_connection is None:
                self._namespace_connection = conn  # type: ignore[assignment]
        elif self._ldb is None:
            self._ldb = conn
        return conn

    def _get_conn(self, uri: str | None) -> DBConnection:
        """Returns the underlying lancedb connection. If the connection
        was initialized using a namespace, this will return the
        LanceNamespaceDBConnection.
        """
        # use the namespace connection if set
        if self.namespace_impl is not None:
            if not self._namespace_connection:
                assert self.namespace_properties is not None, (
                    "namespace_properties must be set when namespace_impl is set"
                )
                namespace_client = namespace_connect(
                    self.namespace_impl, self.namespace_properties
                )

                ns_conn = LanceNamespaceDBConnection(
                    namespace_client,
                    storage_options=self._storage_options,
                    **self._kwargs,
                )
            _LOG.info(f"using namespace connection: {self._namespace_connection=}")
            return ns_conn

        conn = lancedb.connect(
            uri or self.uri,
            region=self._region,
            api_key=self._api_key,
            host_override=self._host_override,
            storage_options=self._storage_options,
            **self._kwargs,
        )
        return conn

    @cached_property
    def _history(self) -> "JobStateManager":  # noqa: F821
        """Returns a JobStateManager that persists job executions and statuses"""
        from geneva.jobs import JobStateManager

        if self._jobs_manager is None:
            self._jobs_manager = JobStateManager(self)
        return self._jobs_manager

    @cached_property
    def flight_client(self) -> Any:
        if self._flight_client is not None:
            return self._flight_client
        import flightsql

        flightsql_client = flightsql.FlightSQLClient
        url = urlparse(self._host_override)
        hostname = url.hostname
        client = flightsql_client(
            host=hostname,
            port=10025,
            token="DATABASE_TOKEN",  # Dummy auth, not plugged in yet
            metadata={"database": self.uri},  # Name of the project-id
            features={"metadata-reflection": "true"},
            insecure=True,  # or False, up to you
        )
        self._flight_client = client
        return client

    @override
    def table_names(
        self, page_token: str | None = None, limit: int | None = None, *args, **kwargs
    ) -> Iterable[str]:
        """List all available tables and views."""
        return self._connect.table_names(
            *args, page_token=page_token, limit=limit or 10, **kwargs
        )

    @override
    def open_table(
        self,
        name: str,
        storage_options: dict[str, str] | None = None,
        index_cache_size: int | None = None,
        version: int | None = None,
        namespace: list[str] | None = None,
        *args,
        **kwargs,
    ) -> "Table":
        """Open a Lance Table.

        Parameters
        ----------
        name: str
            Name of the table.
        storage_options: dict[str, str], optional
            Additional options for the storage backend.
            Options already set on the connection will be inherited by the table,
            but can be overridden here. See available options at
            [https://lancedb.github.io/lancedb/guides/storage/](https://lancedb.github.io/lancedb/guides/storage/)
        namespace: list[str], optional
            Namespace path for the table (e.g., ["workspace"])

        """
        from .table import Table

        storage_options = storage_options or self._storage_options
        namespace = namespace if namespace is not None else []

        return Table(
            self,
            name,
            namespace=namespace,
            index_cache_size=index_cache_size,
            storage_options=storage_options,
            version=version,
        )

    def _validate_existing_table_has_stable_row_ids(self, name: str) -> None:
        """Validate that an existing table has stable row IDs enabled.

        Called when exist_ok=True and stable row IDs are requested, to ensure
        the existing table matches the requested configuration.

        Raises:
            ValueError: If table exists with data but doesn't have stable row IDs
        """
        # Use lancedb table API (not lance.dataset directly) to support namespaces
        if name not in self._connect.table_names():
            return

        try:
            table = self._connect.open_table(name)
            fragments = list(table.to_lance().get_fragments())  # type: ignore[union-attr]
            if fragments:
                if not has_stable_row_ids(fragments):
                    raise ValueError(
                        f"Cannot open table '{name}' with exist_ok=True: "
                        f"table exists but does not have stable row IDs enabled.\n\n"
                        f"You requested stable row IDs, but the existing table "
                        f"was created without them.\n\n"
                        f"Options:\n"
                        f"  1. Drop and recreate the table with stable row IDs\n"
                        f"  2. Remove storage_options if stable row IDs are "
                        f"not required\n"
                        f"  3. Use mode='overwrite' to replace the table"
                    )
            else:
                _LOG.warning(
                    f"Table '{name}' exists but is empty - cannot verify "
                    f"if stable row IDs are enabled."
                )
        except ValueError:
            raise
        except Exception as e:
            _LOG.debug(
                f"Could not check stable row ID status for existing table '{name}': {e}"
            )

    @override
    def create_table(  # type: ignore
        self,
        name: str,
        data: DATA | None = None,
        schema: pa.Schema | LanceModel | None = None,
        mode: str = "create",
        exist_ok: bool = False,
        on_bad_vectors: str = "error",
        fill_value: float = 0.0,
        storage_options: dict[str, str] | None = None,
        *args,
        **kwargs,
    ) -> "Table":  # type: ignore
        """Create a Table in the lake

        Parameters
        ----------
        name: str
            The name of the table
        data: The data to initialize the table, *optional*
            User must provide at least one of `data` or `schema`.
            Acceptable types are:

            - list-of-dict
            - pandas.DataFrame
            - pyarrow.Table or pyarrow.RecordBatch
        schema: The schema of the table, *optional*
            Acceptable types are:

            - pyarrow.Schema
            - lancedb.pydantic.LanceModel
        mode: str; default "create"
            The mode to use when creating the table.
            Can be either "create" or "overwrite".
            By default, if the table already exists, an exception is raised.
            If you want to overwrite the table, use mode="overwrite".
        exist_ok: bool, default False
            If a table by the same name already exists, then raise an exception
            if exist_ok=False. If exist_ok=True, then open the existing table;
            it will not add the provided data but will validate against any
            schema that's specified.
        on_bad_vectors: str, default "error"
            What to do if any of the vectors are not the same size or contain NaNs.
            One of "error", "drop", "fill".
        """
        from .table import Table

        # Handle new_table_enable_stable_row_ids validation and normalization
        # Accept both string "true" and boolean True for user convenience
        if storage_options:
            enable_stable = storage_options.get("new_table_enable_stable_row_ids")
            if enable_stable in ("true", True, 1):
                # Normalize to string "true" for lancedb
                storage_options = dict(storage_options)
                storage_options["new_table_enable_stable_row_ids"] = "true"

                # Validate exist_ok: ensure existing table has stable row IDs
                if exist_ok:
                    self._validate_existing_table_has_stable_row_ids(name)

        # Create table through lancedb
        conn = self._connect

        if self._host_override:
            # in OSS, exist_ok is a separate param, but in phalanx it is set as "mode"
            # workaround until https://github.com/lancedb/lancedb/issues/2900
            if exist_ok and mode == "create":
                mode = "exist_ok"

            if storage_options:
                _LOG.warning(
                    "storage_options parameter is not supported when creating "
                    "tables on remote connections, ignoring"
                )
        else:
            # these params are not supported in remote connections
            kwargs.update(
                exist_ok=exist_ok,
                on_bad_vectors=on_bad_vectors,
                storage_options=storage_options,
            )

        conn.create_table(
            name,
            data,
            schema,
            mode,
            *args,
            fill_value=fill_value,
            **kwargs,
        )
        # Extract namespace from kwargs to pass to Table
        namespace = kwargs.get("namespace")
        return Table(self, name, namespace=namespace, storage_options=storage_options)

    def create_view(
        self,
        name: str,
        query: str,
        materialized: bool = False,
    ) -> "Table":
        """Create a View from a Query.

        Parameters
        ----------
        name: str
            Name of the view.
        query: str
            SQL query to create the view.
        materialized: bool, optional
            If True, the view is materialized.
        """
        if materialized:
            # idea, rename the provided name, and use it as the basis for the
            # materialized view.
            # - how do we add the udfs to the final materialized view table?
            NotImplementedError(
                "creating materialized view via sql query is not supported yet."
            )

        # TODO add test coverage here
        self.sql(f"CREATE VIEW {name} AS ({query})")  # type: ignore[attr-defined]
        return self.open_table(name)

    def create_materialized_view(
        self,
        name: str,
        query: "GenevaQueryBuilder",
        with_no_data: bool = True,
    ) -> "Table":
        """
        Create a materialized view

        Parameters
        ----------
        name: str
            Name of the materialized view.
        query: GenevaQueryBuilder
            Query to create the view.
        with_no_data: bool, optional
            If True, the view is materialized, if false it is ready for refresh.
        """
        from geneva.query import GenevaQueryBuilder

        if not isinstance(query, GenevaQueryBuilder):
            raise ValueError(
                "Materialized views only support plain queries (where, select)"
            )

        tbl = query.create_materialized_view(self, name)
        if not with_no_data and hasattr(tbl, "refresh_view"):
            tbl.refresh_view(name)  # type: ignore[attr-defined]

        return tbl

    def create_udtf_view(
        self,
        name: str,
        source: "GenevaQueryBuilder",
        udtf: "UDTF",
    ) -> "Table":
        """Create a UDTF-backed materialized view.

        .. warning::
            This API is in **beta** and may change in future releases.

        The view is created empty; call ``view.refresh()`` to populate it.

        Parameters
        ----------
        name : str
            Name for the new view table.
        source : GenevaQueryBuilder
            Query defining the source data.
        udtf : UDTF
            The UDTF to execute on refresh.
        """

        from geneva.packager import marshal_udtf
        from geneva.query import (
            MATVIEW_META_BASE_DBURI,
            MATVIEW_META_BASE_TABLE,
            MATVIEW_META_BASE_VERSION,
            MATVIEW_META_NAMESPACE_PATH,
            MATVIEW_META_SOURCE_QUERY,
            MATVIEW_META_UDTF,
            MATVIEW_META_VERSION,
            GenevaQueryBuilder,
        )

        if not isinstance(source, GenevaQueryBuilder):
            raise ValueError("source must be a GenevaQueryBuilder")

        # Validate input_columns against source schema if specified
        # Note: _rowid/_rowaddr are Lance internal columns always available
        if udtf.input_columns:
            src_schema = source._table.schema
            missing = [
                c
                for c in udtf.input_columns
                if c not in src_schema.names and c not in _LANCE_INTERNAL_COLUMNS
            ]
            if missing:
                raise ValueError(
                    f"UDTF input_columns {missing} not found in source schema. "
                    f"Available: {src_schema.names}"
                )

        # Serialize UDTF and source query
        udtf_spec = marshal_udtf(udtf)
        source_query_json = source.to_query_object().model_dump_json()

        # Get source table info
        local_tbl = source._table._native_tbl
        db_uri = _get_db_uri(source._table.uri)

        # Build metadata
        metadata = {
            MATVIEW_META_UDTF: udtf_spec.to_json(),
            MATVIEW_META_SOURCE_QUERY: source_query_json,
            MATVIEW_META_BASE_TABLE: local_tbl.name,
            MATVIEW_META_BASE_DBURI: db_uri,
            MATVIEW_META_BASE_VERSION: str(local_tbl.version),
            MATVIEW_META_VERSION: "udtf",
        }

        # Store namespace path as $-delimited string (if source is in a child namespace)
        if source._table._namespace:
            metadata[MATVIEW_META_NAMESPACE_PATH] = "$".join(source._table._namespace)

        # Create empty table with UDTF output schema + metadata.
        # Delegates to self.create_table() which is namespace-aware.
        schema_with_meta = udtf.output_schema.with_metadata(metadata)
        empty_table = schema_with_meta.empty_table()

        return self.create_table(name, data=empty_table)

    def create_scalar_udtf_view(
        self,
        name: str,
        source: "GenevaQueryBuilder",
        scalar_udtf: "ScalarUDTF",
    ) -> "Table":
        """Create a scalar UDTF-backed materialized view (1:N row expansion).

        .. warning::
            This API is in **beta** and may change in future releases.

        The view is created with placeholder rows (one per source row) that are
        populated on ``view.refresh()``. Each source row expands to zero or more
        output rows via the scalar UDTF.

        Parameters
        ----------
        name : str
            Name for the new view table.
        source : GenevaQueryBuilder
            Query defining the source data.
        scalar_udtf : ScalarUDTF
            The scalar UDTF to execute on refresh.
        """

        from geneva.packager import marshal_scalar_udtf
        from geneva.query import (
            MATVIEW_META_BASE_DBURI,
            MATVIEW_META_BASE_TABLE,
            MATVIEW_META_BASE_VERSION,
            MATVIEW_META_NAMESPACE_PATH,
            MATVIEW_META_SCALAR_UDTF,
            MATVIEW_META_SOURCE_QUERY,
            MATVIEW_META_VERSION,
            MATVIEW_VERSION_SCALAR_UDTF,
            GenevaQueryBuilder,
        )

        if not isinstance(source, GenevaQueryBuilder):
            raise ValueError("source must be a GenevaQueryBuilder")

        # Validate input columns against query projection (not base table)
        if scalar_udtf.input_columns:
            projected = source._schema_for_query(include_metacols=False)
            projected_names = projected.names
            missing = [c for c in scalar_udtf.input_columns if c not in projected_names]
            if missing:
                raise ValueError(
                    f"ScalarUDTF input_columns {missing} not found "
                    f"in query projection. "
                    f"Available: {projected_names}"
                )

        # Serialize scalar UDTF and source query
        spec = marshal_scalar_udtf(scalar_udtf)
        source_query_json = source.to_query_object().model_dump_json()

        # Get source table info
        local_tbl = source._table._native_tbl
        db_uri = _get_db_uri(source._table.uri)

        # Build the output schema: __source_row_id + __child_index
        # + inherited source columns + UDTF output columns
        inherited_schema = source._schema_for_query(include_metacols=False)
        view_fields = [
            pa.field("__source_row_id", pa.int64()),
            pa.field("__child_index", pa.int32()),
        ]
        # Add inherited source columns
        view_fields.extend(
            field
            for field in inherited_schema
            if field.name not in ("__source_row_id", "__child_index", "__is_set")
        )
        # Add UDTF output columns
        view_fields.extend(scalar_udtf.output_schema)

        view_schema = pa.schema(view_fields)

        # Build metadata
        metadata = {
            MATVIEW_META_SCALAR_UDTF: spec.to_json(),
            MATVIEW_META_SOURCE_QUERY: source_query_json,
            MATVIEW_META_BASE_TABLE: local_tbl.name,
            MATVIEW_META_BASE_DBURI: db_uri,
            MATVIEW_META_BASE_VERSION: str(local_tbl.version),
            MATVIEW_META_VERSION: MATVIEW_VERSION_SCALAR_UDTF,
        }

        # Store namespace path as $-delimited string (if source is in a child namespace)
        if source._table._namespace:
            metadata[MATVIEW_META_NAMESPACE_PATH] = "$".join(source._table._namespace)

        # Create empty table with the full schema + metadata
        schema_with_meta = view_schema.with_metadata(metadata)
        empty_table = schema_with_meta.empty_table()

        return self.create_table(name, data=empty_table)

    def drop_view(self, name: str) -> pa.Table:
        """Drop a view."""
        return self.sql(f"DROP VIEW {name}")  # type: ignore[attr-defined]

    @override
    def drop_table(self, name: str, *args, **kwargs) -> None:
        """Drop a table."""
        self._connect.drop_table(name, *args, **kwargs)

    def define_cluster(self, name: str, cluster: "GenevaCluster") -> None:  # noqa: F821
        """
        Define a persistent Geneva cluster. This will upsert the cluster definition by
        name. The cluster can then be provisioned using `context(cluster=name)`.

        Parameters
        ----------
        name: str
            Name of the cluster. This will be used as the key when upserting and
            provisioning the cluster. The cluster name must comply with RFC 12123.
        cluster: GenevaCluster
            The cluster definition to store.
        """
        from geneva.cluster.mgr import ClusterConfigManager

        if self._cluster_manager is None:
            self._cluster_manager = ClusterConfigManager(self)

        cluster.name = name
        self._cluster_manager.upsert(cluster)

    def list_clusters(self) -> list["GenevaCluster"]:  # noqa: F821
        """
        List the cluster definitions. These can be defined using `define_cluster()`.

        Returns
        -------
        list[GenevaCluster]
            List of Geneva cluster definitions
        """
        from geneva.cluster.mgr import ClusterConfigManager

        if self._cluster_manager is None:
            self._cluster_manager = ClusterConfigManager(self)
        return self._cluster_manager.list()

    def delete_cluster(self, name: str) -> None:  # noqa: F821
        """
        Delete a Geneva cluster definition.

        Parameters
        ----------
        name: str
            Name of the cluster to delete.
        """
        from geneva.cluster.mgr import ClusterConfigManager

        if self._cluster_manager is None:
            self._cluster_manager = ClusterConfigManager(self)

        self._cluster_manager.delete(name)

    def define_manifest(
        self,
        name: str,
        manifest: "GenevaManifest",  # noqa: F821
        uploader: Uploader | None = None,
    ) -> None:
        """
        Define a persistent Geneva Manifest that represents the files and dependencies
        used in the execution environment. This will upsert the manifest definition by
        name and upload the required artifacts. The manifest can then be used with
        `context(manifest=name)`.

        Parameters
        ----------
        name: str
            Name of the manifest. This will be used as the key when upserting and
            loading the manifest.
        manifest: GenevaManifest
            The manifest definition to use.
        uploader: Uploader, optional
            An optional, custom Uploader to use. If not provided, the uploader will be
            auto-detected based on the
            environment configuration.
        """
        from geneva.manifest.mgr import ManifestConfigManager

        if self._manifest_manager is None:
            self._manifest_manager = ManifestConfigManager(self)

        # Ensure the manifest table exists before creating Uploader
        # This guarantees the table location can be queried
        _ = self._manifest_manager.get_table()

        # If no uploader is provided, create one with manifest table context
        if uploader is None:
            from geneva.manifest.mgr import MANIFEST_TABLE_NAME

            # If upload_dir is configured, use it directly for manifest uploads
            if self._upload_dir:
                uploader = Uploader(upload_dir=self._upload_dir)
            elif self.namespace_impl and self.namespace_properties:
                # Create uploader for namespace table
                # Build table_id for manifest table
                table_id = self.system_namespace + [MANIFEST_TABLE_NAME]
                uploader = Uploader(
                    namespace_impl=self.namespace_impl,
                    namespace_properties=self.namespace_properties,
                    table_id=table_id,
                )
            else:
                # For local/object-store direct connections, use the
                # "__system" DB URI as the manifest-table root.
                table_id = [MANIFEST_TABLE_NAME]
                uploader = Uploader(
                    db_uri=self.system_table_db_uri,
                    table_id=table_id,
                )

        with upload_local_env(
            # todo: implement excludes
            uploader=uploader,
            zip_output_dir=manifest.local_zip_output_dir,
            delete_local_zips=manifest.delete_local_zips,
            skip_site_packages=manifest.skip_site_packages,
        ) as zips:
            m = copy.deepcopy(manifest)
            m.name = name
            m.zips = zips
            m.checksum = manifest.compute_checksum()
            self._manifest_manager.upsert(m)

    def list_manifests(self) -> list["GenevaManifest"]:  # noqa: F821
        """
        List the manifest definitions. These can be defined using `define_manifest()`.

        Returns
        -------
        list[GenevaManifest]
            List of Geneva manifest definitions
        """
        from geneva.manifest.mgr import ManifestConfigManager

        if self._manifest_manager is None:
            self._manifest_manager = ManifestConfigManager(self)
        return self._manifest_manager.list()

    def delete_manifest(self, name: str) -> None:  # noqa: F821
        """
        Delete a Geneva manifest definition.

        Parameters
        ----------
        name: str
            Name of the manifest to delete.
        """
        from geneva.manifest.mgr import ManifestConfigManager

        if self._manifest_manager is None:
            self._manifest_manager = ManifestConfigManager(self)

        self._manifest_manager.delete(name)

    @staticmethod
    def local_ray_context() -> contextlib.AbstractContextManager[None]:
        """Context manager for a local Ray instance.
        This will provision a local Ray instance and return a context manager.
        This is useful for development or small jobs.
        """
        from geneva.runners.ray._mgr import ray_cluster

        return ray_cluster(local=True, log_to_driver=True, logging_level=logging.INFO)

    def context(
        self,
        cluster: str,
        manifest: str | None = None,
        on_exit=None,
        wait_timeout: float | None = None,
        log_to_driver: bool = True,
        logging_level=logging.INFO,
    ) -> contextlib.AbstractContextManager[None]:
        """Context manager for a Geneva Execution Environment.
            This will provision a cluster based on the cluster
            definition and the manifest provided.
            By default, the context manager will delete the cluster on exit.
            This can be configured with the on_exit parameter.
        Parameters
        ----------
        cluster: str
            Name of the persisted cluster definition to use. Required.
            This will raise an exception if the cluster definition was not
            defined via `define_cluster()`.
        manifest: str, optional
            Optional name of the persisted manifest to use. This will
            raise an exception if the manifest definition was not
            defined via `define_manifest()`. If manifest is not provided,
            the local environment will be uploaded.
        on_exit: ExitMode, optional, default ExitMode.DELETE
            Exit mode for the cluster. By default, the cluster waits for all
            running jobs to complete before deleting.
            To retain the cluster when any job fails or the context body
            raises an exception, use `ExitMode.RETAIN_ON_FAILURE`.
            To always retain the cluster, use `ExitMode.RETAIN`.
        wait_timeout: float, optional, default None
            Internal/experimental. Maximum seconds to wait for tracked jobs
            during context exit. Only applies with DELETE or
            RETAIN_ON_FAILURE. None means wait indefinitely. For
            RETAIN_ON_FAILURE, a timeout is treated as a failure and the
            cluster is retained.
        log_to_driver: bool, optional, default True
            Whether to send Ray worker logs to the driver. Defaults to True for
            better visibility in tests and debugging.
        logging_level: int, optional, default logging.INFO
            The logging level for Ray workers. Use logging.DEBUG for detailed logs.
        """
        from geneva.cluster.mgr import ClusterConfigManager
        from geneva.manifest.mgr import ManifestConfigManager
        from geneva.runners.ray._mgr import ray_cluster
        from geneva.runners.ray.raycluster import ExitMode

        if self._cluster_manager is None:
            self._cluster_manager = ClusterConfigManager(self)
        if self._manifest_manager is None:
            self._manifest_manager = ManifestConfigManager(self)

        if cluster is None:
            raise ValueError(
                "cluster name is required. Use conn.context(cluster=...) or "
                "conn.local_ray_context() for a local Ray instance."
            )

        cluster_def = self._cluster_manager.load(cluster)
        if cluster_def is None:
            raise Exception(
                f"cluster definition '{cluster}' not found. "
                f"Create a new cluster with define_cluster()"
            )

        if cluster_def.cluster_type == GenevaClusterType.LOCAL_RAY:
            if manifest is not None:
                raise ValueError(
                    "Manifests are not supported with LOCAL_RAY cluster type"
                )
            return ray_cluster(
                local=True, log_to_driver=log_to_driver, logging_level=logging_level
            )

        ray_env = {
            "RAY_BACKEND_LOG_LEVEL": "info",
            "RAY_LOG_TO_DRIVER": "1",
            "RAY_ENABLE_RECORD_ACTOR_TASK_LOGGING": "0",
            "RAY_RUNTIME_ENV_LOG_TO_DRIVER_ENABLED": "true",
        }

        # load the manifest if provided
        manifest_def = None
        if manifest is not None:
            manifest_def = self._manifest_manager.load(manifest)
            if manifest_def is None:
                raise Exception(
                    f"manifest definition '{manifest}' not found. "
                    f"Create a new manifest with define_manifest()"
                )

        if cluster_def.cluster_type == GenevaClusterType.EXTERNAL_RAY:
            return ray_cluster(
                addr=cluster_def.ray_address,
                use_portforwarding=False,
                manifest=manifest_def,
                extra_env=ray_env,
                log_to_driver=log_to_driver,
                logging_level=logging_level,
                ray_init_kwargs=cluster_def.ray_init_kwargs,
            )

        use_portforwarding = cluster_def.as_dict()["kuberay"].get(
            "use_portforwarding", True
        )
        rc = cluster_def.to_ray_cluster()
        rc.on_exit = on_exit or ExitMode.DELETE
        if wait_timeout is not None:
            rc.wait_timeout = wait_timeout
        if manifest_def:
            # image explicitly provided in manifest takes precedence over cluster
            if img := manifest_def.head_image:
                _LOG.debug(f"overriding cluster head image from manifest: {img}")
                rc.head_group.image = img
            if img := manifest_def.worker_image:
                _LOG.debug(f"overriding cluster worker image from manifest: {img}")
                for wg in rc.worker_groups:
                    wg.image = img

        return ray_cluster(
            use_portforwarding=use_portforwarding,
            ray_cluster=rc,
            manifest=manifest_def,
            extra_env=ray_env,
            log_to_driver=log_to_driver,
            logging_level=logging_level,
        )

    def is_remote(self) -> bool:
        return self.uri.startswith("db://")


def sql(self, query: str) -> pa.Table:
    """Execute a raw SQL query.

    It uses the Flight SQL engine to execute the query.

    Parameters
    ----------
    query: str
        SQL query to execute

    Returns
    -------
    pyarrow.Table
        Result of the query in a `pyarrow.Table`

    """
    info = self.flight_client.execute(query)
    return self.flight_client.do_get(info.endpoints[0].ticket).read_all()


@attrs.define
class _GenavaConnectionConfig(ConfigBase):
    region: str = attrs.field(default="us-east-1")
    api_key: str | None = attrs.field(default=None)
    host_override: str | None = attrs.field(default=None)
    checkpoint: str | None = attrs.field(default=None)
    system_namespace: list[str] = attrs.field(factory=lambda: [SYSTEM_NAMESPACE])

    @classmethod
    @override
    def name(cls) -> str:
        return "connection"


def _ensure_system_namespace_exists(conn: Connection) -> None:
    """Ensure system_namespace exists, creating it if necessary.

    This is called once at connection time for namespace connections to ensure
    the system namespace is ready before any system tables are created.
    For nested namespaces, creates parent namespaces first.

    Args:
        conn: The Geneva connection

    Raises:
        RuntimeError: If the namespace doesn't exist and cannot be created
    """
    from lance_namespace import CreateNamespaceRequest, DescribeNamespaceRequest

    if not conn.system_namespace or conn._system_namespace_ensured:
        return

    # Remote db:// connections rely on the server to handle namespace creation
    # during table operations. Explicit namespace creation is not reliable across
    # phalanx/server versions and can fail before any useful work begins.
    if conn.namespace_impl is None and conn.is_remote():
        _LOG.info(
            "Deferring system namespace creation for remote connection: %s",
            conn.system_namespace,
        )
        conn._system_namespace_ensured = True
        return

    if conn.namespace_client is None:
        return

    try:
        # Check if namespace exists
        conn.namespace_client.describe_namespace(
            DescribeNamespaceRequest(id=conn.system_namespace)
        )
        _LOG.info(f"System namespace {conn.system_namespace} exists")
        conn._system_namespace_ensured = True
    except Exception as e:
        # Namespace doesn't exist, try to create it
        if "not found" in str(e).lower():
            try:
                # For nested namespaces, create parents first
                for i in range(1, len(conn.system_namespace) + 1):
                    parent_ns = conn.system_namespace[:i]
                    try:
                        conn.namespace_client.describe_namespace(
                            DescribeNamespaceRequest(id=parent_ns)
                        )
                        _LOG.debug(f"Namespace {parent_ns} already exists")
                    except Exception:
                        _LOG.info(f"Creating namespace {parent_ns}...")
                        conn.namespace_client.create_namespace(
                            CreateNamespaceRequest(id=parent_ns)
                        )
                        _LOG.info(f"Created namespace {parent_ns}")

                _LOG.info(f"System namespace {conn.system_namespace} is ready")
                conn._system_namespace_ensured = True
            except Exception as create_error:
                raise RuntimeError(
                    f"Cannot create system namespace {conn.system_namespace}: "
                    f"{create_error}"
                ) from create_error
        else:
            # Some other error, re-raise
            raise


def connect(
    uri: str | Path | None = None,
    *,
    region: str | None = None,
    api_key: Credential | str | None = None,
    host_override: str | None = None,
    storage_options: dict[str, str] | None = None,
    checkpoint: str | CheckpointStore | None = None,
    system_namespace: list[str] | None = None,
    namespace_impl: str | None = None,
    namespace_properties: dict[str, str] | None = None,
    upload_dir: str | None = None,
    initialize_upload_dir: bool = True,
    **kwargs,
) -> Connection:
    """Create a Geneva Connection to an existing database.

    Examples
    --------
        import geneva
        # Connect to a database in object storage
        conn = geneva.connect("s3://my-storage-bucket/my-database")
        # Connect using directory namespace
        conn = geneva.connect(
            namespace_impl="dir", namespace_properties={"root": "/path"}
        )
        # Connect using REST namespace
        conn = geneva.connect(
            namespace_impl="rest", namespace_properties={"uri": f"http://127.0.0.1:1234"}
        )
        # Connect with Phalanx remote server and separate upload bucket
        conn = geneva.connect(
            uri="db://my_database",
            api_key="my-api-key",
            host_override="https://phalanx.example.com",
            upload_dir="s3://my-upload-bucket/manifests",
        )
        tbl = conn.open_table("youtube_dataset")

    Parameters
    ----------
    uri: geneva URI, or Path, optional
        LanceDB Database URI, or a S3/GCS path.
        If not provided and namespace_impl is set, defaults to "namespace://".
    region: str | None
        LanceDB cloud region. Set to `None` on LanceDB Enterprise
    api_key: str | None
        Optional API key for enterprise endpoint
    host_override: str | None
        Optional host URI for enterprise endpoint
    system_namespace: list[str] | None
        Namespace for system tables (manifests, clusters, jobs, errors).
        Defaults to config value if not provided.
    namespace_impl: str | None
        The namespace implementation to use (e.g., "dir", "rest").
        If provided, connects using namespace instead of local database.
    namespace_properties: dict[str, str] | None
        Configuration properties for the namespace implementation.
    upload_dir: str | None
        Optional separate bucket/path for manifest uploads. This allows the client to
        upload manifests to a separate bucket instead of the data bucket
        where manifests are uploaded by default.
    Returns
    -------
    Connection - A LanceDB connection
    """

    api_key, checkpoint_store, host_override, region, uri, system_namespace = (
        _pre_connect(
            api_key,
            checkpoint,
            host_override,
            region,
            uri,
            namespace_impl,
            namespace_properties,
            system_namespace,
        )
    )

    if host_override:
        _LOG.info(f"Using host_override: {host_override}")

    conn = Connection(
        str(uri),
        region=region,
        api_key=api_key,
        host_override=host_override,
        storage_options=storage_options,
        checkpoint_store=checkpoint_store,
        namespace_impl=namespace_impl,
        namespace_properties=namespace_properties,
        system_namespace=system_namespace,
        upload_dir=upload_dir,
        **kwargs,
    )

    if initialize_upload_dir:
        _set_upload_dir(conn, upload_dir)

    # Validate and create system namespace for namespace_impl connections.
    # For remote db:// connections, defer creation until system tables are used.
    if namespace_impl is not None:
        _ensure_system_namespace_exists(conn)

    return conn


def resolve_table_physical_uri(
    table_id: list[str],
    *,
    db_uri: str | None,
    api_key: Credential | str | None = None,
    host_override: str | None = None,
    namespace_impl: str | None = None,
    namespace_properties: dict[str, str] | None = None,
) -> str:
    """Resolve a table reference to the backing physical storage URI."""
    table_name = table_id[-1]
    namespace = table_id[:-1]

    if namespace_impl is not None and namespace_properties is not None:
        from lance_namespace import DescribeTableRequest

        namespace_client = namespace_connect(namespace_impl, namespace_properties)
        response = namespace_client.describe_table(DescribeTableRequest(id=table_id))
        if response.location is None:
            raise ValueError(f"Table {table_id} does not have a physical location")
        return response.location

    if db_uri is not None and db_uri.startswith("db://"):
        remote_db = connect(
            db_uri,
            api_key=api_key,
            host_override=host_override,
            checkpoint="memory",
            initialize_upload_dir=False,
            system_namespace=infer_system_namespace(
                uri=db_uri,
                namespace_impl=namespace_impl,
                system_namespace=None,
            ),
            read_consistency_interval=timedelta(0),
        )
        try:
            table = remote_db.open_table(table_name, namespace=namespace)
            return table.uri  # pyright: ignore[reportAttributeAccessIssue]
        finally:
            remote_db.close()

    assert db_uri is not None, "db_uri must be set"
    return str(URL(str(db_uri)) / f"{table_name}.lance")


def _set_upload_dir(conn: Connection, upload_dir: str | None) -> None:
    # Set up default uploader if not already configured
    # This is needed for cluster operations (like upload_local_env) that don't have
    # table context
    from geneva.config import override_config_kv

    if upload_dir:
        override_config_kv({"uploader.upload_dir": upload_dir})
    else:
        # Remote and namespace-backed connections already have enough table
        # context to build an uploader lazily when needed. Probing the manifest
        # table here can create concurrent side effects during worker-side
        # TableReference reopens.
        if conn.is_remote() or conn.namespace_impl is not None:
            return
        try:
            Uploader.get()
        except (TypeError, ValueError):
            # Uploader not configured - try to derive a default upload_dir
            try:
                from geneva.manifest.mgr import ManifestConfigManager

                mgr = ManifestConfigManager(conn)
                # get the underlying object storage URI
                tbl_uri: str = mgr.table.uri  # pyright: ignore[reportAttributeAccessIssue]
                inner_uri = _get_db_uri(tbl_uri)

                default_upload_dir = f"{inner_uri}/{DEFAULT_UPLOAD_DIR}"
                override_config_kv({"uploader.upload_dir": default_upload_dir})
            except (AttributeError, TypeError, ValueError, OSError, RuntimeError):
                # Can't derive upload_dir - leave unconfigured
                # Uploads will fail later if attempted without explicit upload_dir
                pass


def _pre_connect(
    api_key: Credential | str | None,
    checkpoint: str | CheckpointStore | None,
    host_override: str | None,
    region: str | None,
    uri: str | Path | None,
    namespace_impl: str | None,
    namespace_properties: dict[str, str] | None,
    system_namespace: list[str] | None,
) -> tuple[
    Credential | None,
    CheckpointStore,
    str | None,
    str,
    str | Path,
    list[str],
]:
    # load values from config if not provided via arguments
    config = _GenavaConnectionConfig.get()
    region = region or config.region or "us-east-1"
    api_key = api_key or config.api_key
    api_key = Credential(api_key) if isinstance(api_key, str) else api_key
    host_override = host_override or config.host_override
    if system_namespace is None:
        system_namespace = config.system_namespace
    # handle local relative paths
    is_local = isinstance(uri, Path) or (
        uri is not None and get_uri_scheme(uri) == "file"
    )
    if is_local and uri:
        if isinstance(uri, str):
            uri = Path(uri)
        uri = uri.expanduser().absolute()
        uri.mkdir(parents=True, exist_ok=True)

    # Default URI for namespace connections
    if uri is None:
        uri = "namespace://"

    if checkpoint is None:
        checkpoint = str(URL(str(uri)) / "ckp")
    if isinstance(checkpoint, str):
        checkpoint_store = CheckpointStore.from_uri(checkpoint)
    else:
        checkpoint_store = checkpoint

    _LOG.debug(f"using checkpoint store: {type(checkpoint_store)}")

    return api_key, checkpoint_store, host_override, region, uri, system_namespace


def _get_db_uri(tbl_uri: str) -> str:
    """Get the database URI from a table URI.

    For example,
        s3://bucket/path/to/table.lance -> s3://bucket/path/to
        db://path/to/table.lance -> db://path/to
        /tmp/foo/table.lance -> /tmp/foo
        s3+ddb://bucket/path/table.lance?ddbTableName=x -> s3+ddb://bucket/path?ddbTableName=x
    """
    parsed = urlparse(tbl_uri)
    path_parts = parsed.path.rsplit("/", 1)
    db_path = path_parts[0] if len(path_parts) > 1 else ""

    # Handle local paths (no scheme)
    if not parsed.scheme:
        return db_path

    base = f"{parsed.scheme}://{parsed.netloc}{db_path}"
    if parsed.query:
        base = f"{base}?{parsed.query}"
    return base
